# copyright 2003-2024 LOGILAB S.A. (Paris, FRANCE), all rights reserved.
# contact https://www.logilab.fr/ -- mailto:contact@logilab.fr
#
# This file is part of CubicWeb.
#
# CubicWeb is free software: you can redistribute it and/or modify it under the
# terms of the GNU Lesser General Public License as published by the Free
# Software Foundation, either version 2.1 of the License, or (at your option)
# any later version.
#
# CubicWeb is distributed in the hope that it will be useful, but WITHOUT
# ANY WARRANTY; without even the implied warranty of MERCHANTABILITY or FITNESS
# FOR A PARTICULAR PURPOSE.  See the GNU Lesser General Public License for more
# details.
#
# You should have received a copy of the GNU Lesser General Public License along
# with CubicWeb.  If not, see <https://www.gnu.org/licenses/>.
"""Base class for entity objects manipulated in clients"""

from logging import getLogger
from typing import Tuple, Iterable

from logilab.common.decorators import cached
from logilab.common.registry import yes
from logilab.mtconverter import TransformData, xml_escape
from rql.nodes import Not, VariableRef, Constant, make_relation, Relation as RqlRelation
from rql.stmts import Select
from rql.utils import rqlvar_maker

from cubicweb import REVERSE_INDICATOR, Unauthorized, neg_role, set_log_methods
from cubicweb.appobject import AppObject
from cubicweb.mttransforms import ENGINE
from cubicweb.rqlrewrite import RQLRewriter
from cubicweb.rset import ResultSet
from cubicweb.schema import RQLVocabularyConstraint, RQLConstraint, GeneratedConstraint
from cubicweb.uilib import soup2xhtml


_marker = object()


def greater_card(rschema, subjtypes, objtypes, index):
    for subjtype in subjtypes:
        for objtype in objtypes:
            card = rschema.relation_definition(subjtype, objtype).cardinality[index]
            if card in "+*":
                return card
    return "1"


def can_use_rest_path(value):
    """return True if value can be used at the end of a Rest URL path"""
    if value is None:
        return False

    # make sure value is a string or has a valid string representation
    value = str(value)
    # the check for ?, /, & are to prevent problems when running
    # behind Apache mod_proxy
    if value == "" or "?" in value or "/" in value or "&" in value:
        return False
    return True


def rel_vars(rel):
    return (
        (isinstance(rel.children[0], VariableRef) and rel.children[0].variable or None),
        (
            isinstance(rel.children[1].children[0], VariableRef)
            and rel.children[1].children[0].variable
            or None
        ),
    )


def rel_matches(rel, rtype, role, varname, operator="="):
    if rel.r_type == rtype and rel.children[1].operator == operator:
        same_role_var_idx = 0 if role == "subject" else 1
        variables = rel_vars(rel)
        if variables[same_role_var_idx].name == varname:
            return variables[1 - same_role_var_idx]


def build_cstr_with_linkto_infos(cstr, args, searchedvar, evar, lt_infos, eidvars):
    """restrict vocabulary as much as possible in entity creation,
    based on infos provided by __linkto form param.

    Example based on following schema:

      class works_in(RelationDefinition):
          subject = 'CWUser'
          object = 'Lab'
          cardinality = '1*'
          constraints = [RQLConstraint('S in_group G, O welcomes G')]

      class welcomes(RelationDefinition):
          subject = 'Lab'
          object = 'CWGroup'

    If you create a CWUser in the "scientists" CWGroup you can show
    only the labs that welcome them using :

      lt_infos = {('in_group', 'subject'): 321}

    You get following restriction : 'O welcomes G, G eid 321'

    """
    st = cstr.snippet_rqlst.copy()
    # replace relations in ST by eid infos from linkto where possible
    for (info_rtype, info_role), eids in lt_infos.items():
        eid = eids[0]  # NOTE: we currently assume a pruned lt_info with only 1 eid
        for rel in st.iget_nodes(RqlRelation):
            targetvar = rel_matches(rel, info_rtype, info_role, evar.name)
            if targetvar is not None:
                if targetvar.name in eidvars:
                    rel.parent.remove(rel)
                else:
                    eidrel = make_relation(
                        targetvar, "eid", (targetvar.name, "Substitute"), Constant
                    )
                    rel.parent.replace(rel, eidrel)
                    args[targetvar.name] = eid
                    eidvars.add(targetvar.name)
    # if modified ST still contains evar references we must discard the
    # constraint, otherwise evar is unknown in the final rql query which can
    # lead to a SQL table cartesian product and multiple occurences of solutions
    evarname = evar.name
    for rel in st.iget_nodes(RqlRelation):
        for variable in rel_vars(rel):
            if variable and evarname == variable.name:
                return
    # else insert snippets into the global tree
    return GeneratedConstraint(st, cstr.mainvars - set(evarname))


def pruned_lt_info(eschema, lt_infos):
    pruned = {}
    for (lt_rtype, lt_role), eids in lt_infos.items():
        # we can only use lt_infos describing relation with a cardinality
        # of value 1 towards the linked entity
        if not len(eids) == 1:
            continue
        lt_card = eschema.relation_definition(lt_rtype, lt_role).cardinality[
            0 if lt_role == "subject" else 1
        ]
        if lt_card not in "?1":
            continue
        pruned[(lt_rtype, lt_role)] = eids
    return pruned


def get_attribute_name_with_its_role(attribute_name: str) -> Tuple[str, str]:
    assert isinstance(attribute_name, str)
    assert len(attribute_name) > 0
    if attribute_name.startswith(REVERSE_INDICATOR):
        return (
            attribute_name[len(REVERSE_INDICATOR) :],
            "object",
        )
    return (attribute_name, "subject")


class Entity(AppObject):
    """an entity instance has e_schema automagically set on
    the class and instances has access to their issuing cursor.

    A property is set for each attribute and relation on each entity's type
    class. Becare that among attributes, 'eid' is *NEVER* stored in the
    dict containment (which acts as a cache for other attributes dynamically
    fetched)

    :type e_schema: `cubicweb.schema.EntitySchema`
    :ivar e_schema: the entity's schema

    :type rest_attr: str
    :cvar rest_attr: indicates which attribute should be used to build REST urls
       If `None` is specified (the default), the first unique attribute will
       be used ('eid' if none found)

    :type cw_skip_copy_for: list
    :cvar cw_skip_copy_for: a list of couples (rtype, role) for each relation
       that should be skipped when copying this kind of entity. Note that some
       relations such as composite relations or relations that have '?1' as
       object cardinality are always skipped.
    """

    __registry__ = "etypes"
    __select__ = yes()

    # class attributes that must be set in class definition
    rest_attr = None
    fetch_attrs = None
    cw_skip_copy_for = [("in_state", "subject")]
    # class attributes set automatically at registration time
    e_schema = None

    @classmethod
    def __initialize__(cls, schema):
        """initialize a specific entity class by adding descriptors to access
        entity type's attributes and relations
        """
        etype = cls.__regid__
        assert etype != "Any", etype
        cls.e_schema = eschema = schema.entity_schema_for(etype)
        for rschema, _ in eschema.attribute_definitions():
            if rschema.type == "eid":
                continue
            setattr(cls, rschema.type, Attribute(rschema.type))
        for rschema, _, role in eschema.relation_definitions():
            if role == "subject":
                attr = rschema.type
            else:
                attr = f"{REVERSE_INDICATOR:s}{rschema.type:s}"
            setattr(cls, attr, Relation(rschema, role))

    fetch_attrs = ("modification_date",)

    @classmethod
    def cw_fetch_order(cls, select, attr, var):
        """This class method may be used to control sort order when multiple
        entities of this type are fetched through ORM methods. Its arguments
        are:

        * `select`, the RQL syntax tree

        * `attr`, the attribute being watched

        * `var`, the variable through which this attribute's value may be
          accessed in the query

        When you want to do some sorting on the given attribute, you should
        modify the syntax tree accordingly. For instance:

        .. sourcecode:: python

          from rql import nodes

          class Version(AnyEntity):
              __regid__ = 'Version'

              fetch_attrs = ('num', 'description', 'in_state')

              @classmethod
              def cw_fetch_order(cls, select, attr, var):
                  if attr == 'num':
                      func = nodes.Function('version_sort_value')
                      func.append(nodes.variable_ref(var))
                      sterm = nodes.SortTerm(func, asc=False)
                      select.add_sort_term(sterm)

        The default implementation call
        :meth:`~cubicweb.entity.Entity.cw_fetch_unrelated_order`
        """
        cls.cw_fetch_unrelated_order(select, attr, var)

    @classmethod
    def cw_fetch_unrelated_order(cls, select, attr, var):
        """This class method may be used to control sort order when multiple entities of
        this type are fetched to use in edition (e.g. propose them to create a
        new relation on an edited entity).

        See :meth:`~cubicweb.entity.Entity.cw_fetch_unrelated_order` for a
        description of its arguments and usage.

        By default entities will be listed on their modification date descending,
        i.e. you'll get entities recently modified first.
        """
        if attr == "modification_date":
            select.add_sort_var(var, asc=False)

    @classmethod
    def fetch_rql(
        cls, user, fetchattrs=None, mainvar="X", settype=True, ordermethod="fetch_order"
    ):
        st = cls.fetch_rqlst(
            user,
            mainvar=mainvar,
            fetchattrs=fetchattrs,
            settype=settype,
            ordermethod=ordermethod,
        )
        return st.as_string()

    @classmethod
    def fetch_rqlst(
        cls,
        user,
        select=None,
        mainvar="X",
        fetchattrs=None,
        settype=True,
        ordermethod="fetch_order",
    ):
        if select is None:
            select = Select()
            mainvar = select.get_variable(mainvar)
            select.add_selected(mainvar)
        elif isinstance(mainvar, str):
            assert mainvar in select.defined_vars
            mainvar = select.get_variable(mainvar)
        # eases string -> syntax tree test transition: please remove once stable
        select._varmaker = rqlvar_maker(
            defined=select.defined_vars, aliases=select.aliases, index=26
        )
        if settype:
            rel = select.add_type_restriction(mainvar, cls.__regid__)
            # should use 'is_instance_of' instead of 'is' so we retrieve
            # subclasses instances as well
            rel.r_type = "is_instance_of"
        if fetchattrs is None:
            fetchattrs = cls.fetch_attrs
        cls._fetch_restrictions(mainvar, select, fetchattrs, user, ordermethod)
        return select

    @classmethod
    def _fetch_ambiguous_rtypes(cls, select, var, fetchattrs, subjtypes, schema):
        """find rtypes in `fetchattrs` that relate different subject etypes
        taken from (`subjtypes`) to different target etypes; these so called
        "ambiguous" relations, are added directly to the `select` syntax tree
        selection but removed from `fetchattrs` to avoid the fetch recursion
        because we have to choose only one targettype for the recursion and
        adding its own fetch attrs to the selection -when we recurse- would
        filter out the other possible target types from the result set
        """
        for attr in fetchattrs.copy():
            rschema = schema.relation_schema_for(attr)
            if rschema.final:
                continue
            ttypes = None
            for subjtype in subjtypes:
                cur_ttypes = set(rschema.objects(subjtype))
                if ttypes is None:
                    ttypes = cur_ttypes
                elif cur_ttypes != ttypes:
                    # we found an ambiguous relation: remove it from fetchattrs
                    fetchattrs.remove(attr)
                    # ... and add it to the selection
                    targetvar = select.make_variable()
                    select.add_selected(targetvar)
                    rel = make_relation(var, attr, (targetvar,), VariableRef)
                    select.add_restriction(rel)
                    break

    @classmethod
    def _fetch_restrictions(
        cls, mainvar, select, fetchattrs, user, ordermethod="fetch_order", visited=None
    ):
        eschema = cls.e_schema
        if visited is None:
            visited = set((eschema.type,))
        elif eschema.type in visited:
            # avoid infinite recursion
            return
        else:
            visited.add(eschema.type)
        for attr in sorted(fetchattrs):
            try:
                rschema = eschema.subject_relations[attr]
            except KeyError:
                cls.warning(
                    "skipping fetch_attr %s defined in %s (not found in schema)",
                    attr,
                    cls.__regid__,
                )
                continue
            # XXX take_first=True to remove warning triggered by ambiguous inlined relations
            rdef = eschema.relation_definition(attr, take_first=True)
            if not user.matching_groups(rdef.get_groups("read")):
                continue
            if rschema.final or rdef.cardinality[0] in "?1":
                var = select.make_variable()
                select.add_selected(var)
                rel = make_relation(mainvar, attr, (var,), VariableRef)
                select.add_restriction(rel)
            else:
                cls.warning(
                    "bad relation %s specified in fetch attrs for %s", attr, cls
                )
                continue
            if not rschema.final:
                # XXX we need outer join in case the relation is not mandatory
                # (card == '?')  *or if the entity is being added*, since in
                # that case the relation may still be missing. As we miss this
                # later information here, systematically add it.
                rel.change_optional("right")
                targettypes = rschema.objects(eschema.type)
                vreg = user._cw.vreg  # XXX user._cw.vreg iiiirk
                etypecls = vreg["etypes"].etype_class(targettypes[0])
                if len(targettypes) > 1:
                    # find fetch_attrs common to all destination types
                    fetchattrs = vreg["etypes"].fetch_attrs(targettypes)
                    # ... and handle ambiguous relations
                    cls._fetch_ambiguous_rtypes(
                        select, var, fetchattrs, targettypes, vreg.schema
                    )
                else:
                    fetchattrs = etypecls.fetch_attrs
                etypecls._fetch_restrictions(
                    var, select, fetchattrs, user, None, visited=visited
                )
            if ordermethod is not None:
                cmeth = getattr(cls, "cw_" + ordermethod)
                cmeth(select, attr, var)

    @classmethod
    @cached
    def cw_rest_attr_info(cls):
        """this class method return an attribute name to be used in URL for
        entities of this type and a boolean flag telling if its value should be
        checked for uniqness.

        The attribute returned is, in order of priority:

        * class's `rest_attr` class attribute
        * an attribute defined as unique in the class'schema
        * 'eid'
        """
        mainattr, needcheck = "eid", True
        if cls.rest_attr:
            mainattr = cls.rest_attr
            needcheck = not cls.e_schema.has_unique_values(mainattr)
        else:
            for rschema in list(cls.e_schema.subject_relations.values()):
                if (
                    rschema.final
                    and rschema not in ("eid", "cwuri")
                    and cls.e_schema.has_unique_values(rschema)
                    and cls.e_schema.relation_definition(rschema.type).cardinality[0]
                    == "1"
                ):
                    mainattr = str(rschema)
                    needcheck = False
                    break
        if mainattr == "eid":
            needcheck = False
        return mainattr, needcheck

    @classmethod
    def _cw_build_entity_query(cls, kwargs):
        relations = []
        restrictions = set()
        pendingrels = []
        eschema = cls.e_schema
        qargs = {}
        attrcache = {}
        for attr, value in kwargs.items():
            attr, role = get_attribute_name_with_its_role(attr)
            assert eschema.has_relation(attr, role), "%s %s not found on %s" % (
                attr,
                role,
                eschema,
            )
            rschema = (
                eschema.subject_relations[attr]
                if role == "subject"
                else eschema.object_relations[attr]
            )
            if not rschema.final and isinstance(value, (tuple, list, set, frozenset)):
                if len(value) == 0:
                    continue  # avoid crash with empty IN clause
                elif len(value) == 1:
                    value = next(iter(value))
                else:
                    # prepare IN clause
                    pendingrels.append((attr, role, value))
                    continue
            if rschema.final:  # attribute
                relations.append(f"X {attr} %({attr})s")
                attrcache[attr] = value
            elif value is None:
                pendingrels.append((attr, role, value))
            else:
                rvar = attr.upper()
                if role == "object":
                    relations.append(f"{rvar} {attr} X")
                else:
                    relations.append(f"X {attr} {rvar}")
                restriction = f"{rvar} eid %({attr})s"
                if restriction not in restrictions:
                    restrictions.add(restriction)
                if hasattr(value, "eid"):
                    value = value.eid
            qargs[attr] = value
        rql = ""
        if relations:
            rql += ", ".join(relations)
        if restrictions:
            rql += f" WHERE {', '.join(restrictions)}"
        return rql, qargs, pendingrels, attrcache

    @classmethod
    def _cw_handle_pending_relations(cls, eid, pendingrels, execute):
        for attr, role, values in pendingrels:
            if role == "object":
                restr = f"Y {attr} X"
            else:
                restr = f"X {attr} Y"
            if values is None:
                execute("DELETE %s WHERE X eid %%(x)s" % restr, {"x": eid})
                continue
            execute(
                "SET %s WHERE X eid %%(x)s, Y eid IN (%s)"
                % (restr, ",".join(str(getattr(r, "eid", r)) for r in values)),
                {"x": eid},
                build_descr=False,
            )

    @classmethod
    def cw_instantiate(cls, execute, **kwargs):
        """add a new entity of this given type

        Example (in a shell session):

        >>> companycls = vreg['etypes'].etype_class('Company')
        >>> personcls = vreg['etypes'].etype_class('Person')
        >>> c = companycls.cw_instantiate(session.execute, name=u'Logilab')
        >>> p = personcls.cw_instantiate(session.execute, firstname=u'John', lastname=u'Doe',
        ...                              works_for=c)

        You can also set relations where the entity has 'object' role by
        prefixing the relation name by 'reverse_'. Also, relation values may be
        an entity or eid, a list of entities or eids.
        """
        rql, qargs, pendingrels, attrcache = cls._cw_build_entity_query(kwargs)
        assert isinstance(rql, str)
        entity_type = cls.__regid__
        assert entity_type
        if rql:
            rql = f"INSERT {entity_type} X: {rql}"
        else:
            rql = f"INSERT {entity_type} X"

        try:
            rset = execute(rql, qargs)
            assert len(rset) == 1
            created = rset.get_entity(0, 0)
        except IndexError as error:
            raise Exception(
                f"could not create a {entity_type!r}"
                f" with {rql!r} ({qargs!r})"
                f": {str(error)}"
            )
        created._cw_update_attr_cache(attrcache)
        cls._cw_handle_pending_relations(created.eid, pendingrels, execute)
        return created

    def __init__(self, req, rset=None, row=None, col=0):
        AppObject.__init__(self, req, rset=rset, row=row, col=col)
        self._cw_related_cache = {}
        self._cw_adapters_cache = {}
        if rset is not None:
            self.eid = rset[row][col]
        else:
            self.eid = None
        self._cw_is_saved = True
        self.cw_attr_cache = {}

    def __repr__(self):
        return "<Entity %s %s %s at %s>" % (
            self.e_schema,
            self.eid,
            list(self.cw_attr_cache),
            id(self),
        )

    def __lt__(self, other):
        return NotImplemented

    def __eq__(self, other):
        if isinstance(self.eid, int):
            return self.eid == other.eid
        return self is other

    def __hash__(self):
        if isinstance(self.eid, int):
            return self.eid
        return super(Entity, self).__hash__()

    def _cw_update_attr_cache(self, attrcache):
        trdata = self._cw.transaction_data
        uncached_attrs = trdata.get(f"{self.eid}.storage-special-process-attrs", set())
        uncached_attrs.update(trdata.get(f"{self.eid}.dont-cache-attrs", set()))
        for attr in uncached_attrs:
            attrcache.pop(attr, None)
            self.cw_attr_cache.pop(attr, None)
        self.cw_attr_cache.update(attrcache)

    def _cw_dont_cache_attribute(self, attr, repo_side=False):
        """Called when some attribute has been transformed by a *storage*,
        hence the original value should not be cached **by anyone**.

        For example we have a special "fs_importing" mode in BFSS
        where a file path is given as attribute value and stored as is
        in the data base. Later access to the attribute will provide
        the content of the file at the specified path. We do not want
        the "filepath" value to be cached.

        """
        trdata = self._cw.transaction_data
        trdata.setdefault(f"{self.eid}.dont-cache-attrs", set()).add(attr)
        if repo_side:
            trdata.setdefault(f"{self.eid}.storage-special-process-attrs", set()).add(
                attr
            )

    def __json_encode__(self):
        """custom json dumps hook to dump the entity's eid
        which is not part of dict structure itself
        """
        dumpable = self.cw_attr_cache.copy()
        dumpable["eid"] = self.eid
        return dumpable

    def cw_adapt_to(self, interface):
        """return an adapter the entity to the given interface name.

        return None if it can not be adapted.
        """
        cache = self._cw_adapters_cache
        try:
            return cache[interface]
        except KeyError:
            adapter = self._cw.vreg["adapters"].select_or_none(
                interface, self._cw, entity=self
            )
            cache[interface] = adapter
            return adapter

    def has_eid(self):  # XXX cw_has_eid
        """return True if the entity has an attributed eid (False
        meaning that the entity has to be created
        """
        try:
            int(self.eid)
            return True
        except (ValueError, TypeError):
            return False

    def cw_is_saved(self):
        """during entity creation, there is some time during which the entity
        has an eid attributed though it's not saved (eg during
        'before_add_entity' hooks). You can use this method to ensure the entity
        has an eid *and* is saved in its source.
        """
        return self.has_eid() and self._cw_is_saved

    def cw_check_perm(self, action):
        self.e_schema.check_perm(self._cw, action, eid=self.eid)

    def cw_has_perm(self, action):
        return self.e_schema.has_perm(self._cw, action, eid=self.eid)

    def view(
        self, __vid, __registry="views", w=None, initargs=None, **kwargs
    ):  # XXX cw_view
        """shortcut to apply a view on this entity"""
        if initargs is None:
            initargs = kwargs
        else:
            initargs.update(kwargs)
        view = self._cw.vreg[__registry].select(
            __vid,
            self._cw,
            rset=self.cw_rset,
            row=self.cw_row,
            col=self.cw_col,
            **initargs,
        )
        return view.render(row=self.cw_row, col=self.cw_col, w=w, **kwargs)

    def absolute_url(self, *args, **kwargs):  # XXX cw_url
        """return an absolute url to view this entity"""
        # use *args since we don't want first argument to be "anonymous" to
        # avoid potential clash with kwargs
        if args:
            assert len(args) == 1, "only 0 or 1 non-named-argument expected"
            method = args[0]
        else:
            method = None
        if method in (None, "view"):
            kwargs["_restpath"] = self.rest_path()
        else:
            kwargs["rql"] = f"Any X WHERE X eid {self.eid}"
        return self._cw.build_url(method, **kwargs)

    def rest_path(self):  # XXX cw_rest_path
        """returns a REST-like (relative) path for this entity"""
        mainattr, needcheck = self.cw_rest_attr_info()
        etype = str(self.e_schema)
        path = etype.lower()
        if mainattr != "eid":
            value = getattr(self, mainattr)
            if not can_use_rest_path(value):
                mainattr = "eid"
                path = None
            elif needcheck:
                # make sure url is not ambiguous
                try:
                    nbresults = self.__unique
                except AttributeError:
                    rql = f"Any COUNT(X) WHERE X is {etype}, X {mainattr} %(value)s"
                    nbresults = self.__unique = self._cw.execute(rql, {"value": value})[
                        0
                    ][0]
                if nbresults != 1:  # ambiguity?
                    mainattr = "eid"
                    path = None
        if mainattr == "eid":
            value = self.eid
        if path is None:
            # fallback url: <base-url>/<eid> url is used as cw entities uri,
            # prefer it to <base-url>/<etype>/eid/<eid>
            return str(value)
        return f"{path}/{self._cw.url_quote(value)}"

    def cw_attr_metadata(self, attr, metadata):
        """return a metadata for an attribute (None if unspecified)"""
        value = getattr(self, f"{attr}_{metadata}", None)
        if value is None and metadata == "encoding":
            value = self._cw.vreg.property_value("ui.encoding")
        return value

    def printable_value(
        self, attr, value=_marker, attrtype=None, format="text/html", displaytime=True
    ):  # XXX cw_printable_value
        """return a displayable value (i.e. unicode string) which may contains
        html tags
        """
        attr = str(attr)
        if value is _marker:
            value = getattr(self, attr)
        if isinstance(value, str):
            value = value.strip()
        if value is None or value == "":  # don't use "not", 0 is an acceptable value
            return ""
        if attrtype is None:
            attrtype = self.e_schema.destination(attr)
        props = self.e_schema.relation_definition(attr)
        if attrtype == "String":
            # internalinalized *and* formatted string such as schema
            # description...
            if props.internationalizable:
                value = self._cw._(value)
            attrformat = self.cw_attr_metadata(attr, "format")
            if attrformat:
                return self._cw_mtc_transform(
                    value, attrformat, format, self._cw.encoding
                )
        elif attrtype == "Bytes":
            attrformat = self.cw_attr_metadata(attr, "format")
            if attrformat:
                encoding = self.cw_attr_metadata(attr, "encoding")
                return self._cw_mtc_transform(
                    value.getvalue(), attrformat, format, encoding
                )
            return ""
        value = self._cw.printable_value(
            attrtype, value, props, displaytime=displaytime
        )
        if format == "text/html":
            value = xml_escape(value)
        return value

    def _cw_mtc_transform(self, data, format, target_format, encoding, _engine=ENGINE):
        trdata = TransformData(data, format, encoding, appobject=self)
        data = _engine.convert(trdata, target_format).decode()
        if target_format == "text/html":
            data = soup2xhtml(data, self._cw.encoding)
        return data

    # entity cloning ##########################################################

    def copy_relations(self, ceid):  # XXX cw_copy_relations
        """copy relations of the object with the given eid on this
        object (this method is called on the newly created copy, and
        ceid designates the original entity).

        By default meta and composite relations are skipped.
        Overrides this if you want another behaviour
        """
        assert self.has_eid()
        execute = self._cw.execute
        skip_copy_for = {"subject": set(), "object": set()}
        for rtype, role in self.cw_skip_copy_for:
            assert role in ("subject", "object"), role
            skip_copy_for[role].add(rtype)
        for rschema in list(self.e_schema.subject_relations.values()):
            if rschema.type in skip_copy_for["subject"]:
                continue
            if rschema.final or rschema.meta or rschema.rule:
                continue
            # skip already defined relations
            if getattr(self, rschema.type):
                continue
            # XXX take_first=True to remove warning triggered by ambiguous relations
            rdef = self.e_schema.relation_definition(rschema, take_first=True)
            # skip composite relation
            if rdef.composite:
                continue
            # skip relation with card in ?1 else we either change the copied
            # object (inlined relation) or inserting some inconsistency
            if rdef.cardinality[1] in "?1":
                continue
            rql = "SET X %s V WHERE X eid %%(x)s, Y eid %%(y)s, Y %s V" % (
                rschema.type,
                rschema.type,
            )
            execute(rql, {"x": self.eid, "y": ceid})
            self.cw_clear_relation_cache(rschema.type, "subject")
        for rschema in list(self.e_schema.object_relations.values()):
            if rschema.meta or rschema.rule:
                continue
            # skip already defined relations
            if self.related(rschema.type, "object"):
                continue
            if rschema.type in skip_copy_for["object"]:
                continue
            # XXX take_first=True to remove warning triggered by ambiguous relations
            rdef = self.e_schema.relation_definition(rschema, "object", take_first=True)
            # skip composite relation
            if rdef.composite:
                continue
            # skip relation with card in ?1 else we either change the copied
            # object (inlined relation) or inserting some inconsistency
            if rdef.cardinality[0] in "?1":
                continue
            rql = "SET V %s X WHERE X eid %%(x)s, Y eid %%(y)s, V %s Y" % (
                rschema.type,
                rschema.type,
            )
            execute(rql, {"x": self.eid, "y": ceid})
            self.cw_clear_relation_cache(rschema.type, "object")

    # data fetching methods ###################################################

    @cached
    def as_rset(self):  # XXX .cw_as_rset
        """returns a resultset containing `self` information"""
        rset = ResultSet(
            [(self.eid,)],
            "Any X WHERE X eid %(x)s",
            {"x": self.eid},
            [(self.cw_etype,)],
        )
        rset.req = self._cw
        return rset

    def _cw_to_complete_relations(self):
        """by default complete final relations to when calling .complete()"""
        for rschema in list(self.e_schema.subject_relations.values()):
            if rschema.final:
                continue
            targets = rschema.objects(self.e_schema)
            if rschema.inlined:
                matching_groups = self._cw.user.matching_groups
                if all(
                    matching_groups(e.get_groups("read"))
                    and rschema.relation_definition(self.e_schema, e).get_groups("read")
                    for e in targets
                ):
                    yield rschema, "subject"

    def _cw_to_complete_attributes(self, skip_bytes=True, skip_pwd=True):
        for rschema, attrschema in self.e_schema.attribute_definitions():
            # skip binary data by default
            if skip_bytes and attrschema.type == "Bytes":
                continue
            attr = rschema.type
            if attr == "eid":
                continue
            # password retrieval is blocked at the repository server level
            rdef = rschema.relation_definition(self.e_schema, attrschema)
            if not self._cw.user.matching_groups(rdef.get_groups("read")) or (
                attrschema.type == "Password" and skip_pwd
            ):
                self.cw_attr_cache[attr] = None
                continue
            yield attr

    _cw_completed = False

    def complete(
        self, attributes=None, skip_bytes=True, skip_pwd=True
    ):  # XXX cw_complete
        """complete this entity by adding missing attributes (i.e. query the
        repository to fill the entity)

        :type skip_bytes: bool
        :param skip_bytes:
          if true, attribute of type Bytes won't be considered
        """
        assert self.has_eid()
        if self._cw_completed:
            return
        if attributes is None:
            self._cw_completed = True
        varmaker = rqlvar_maker()
        V = next(varmaker)
        rql = ["WHERE %s eid %%(x)s" % V]
        selected = []
        for attr in attributes or self._cw_to_complete_attributes(skip_bytes, skip_pwd):
            # if attribute already in entity, nothing to do
            if attr in self.cw_attr_cache:
                continue
            # case where attribute must be completed, but is not yet in entity
            var = next(varmaker)
            rql.append(f"{V} {attr} {var}")
            selected.append((attr, var))
        # +1 since this doesn't include the main variable
        lastattr = len(selected) + 1
        if attributes is None:
            # fetch additional relations (restricted to 0..1 relations)
            for rschema, role in self._cw_to_complete_relations():
                rtype = rschema.type
                if self.cw_relation_cached(rtype, role):
                    continue
                # at this point we suppose that:
                # * this is a inlined relation
                # * entity (self) is the subject
                # * user has read perm on the relation and on the target entity
                assert rschema.inlined
                assert role == "subject"
                var = next(varmaker)
                # keep outer join anyway, we don't want .complete to crash on
                # missing mandatory relation (see #1058267)
                rql.append(f"{V} {rtype} {var}?")
                selected.append(((rtype, role), var))
        if selected:
            # select V, we need it as the left most selected variable
            # if some outer join are included to fetch inlined relations
            rql = f"Any {V},{','.join(var for attr, var in selected)} {','.join(rql)}"
            try:
                rset = self._cw.execute(rql, {"x": self.eid}, build_descr=False)[0]
            except IndexError:
                raise Exception(
                    f"unable to fetch attributes for entity with eid {self.eid}"
                )
            # handle attributes
            for i in range(1, lastattr):
                self.cw_attr_cache[str(selected[i - 1][0])] = rset[i]
            # handle relations
            for i in range(lastattr, len(rset)):
                rtype, role = selected[i - 1][0]
                value = rset[i]
                if value is None:
                    rrset = ResultSet([], rql, {"x": self.eid})
                    rrset.req = self._cw
                else:
                    rrset = self._cw.eid_rset(value)
                self.cw_set_relation_cache(rtype, role, rrset)

    def cw_attr_value(self, name):
        """get value for the attribute relation <name>, query the repository
        to get the value if necessary.

        :type name: str
        :param name: name of the attribute to get
        """
        try:
            return self.cw_attr_cache[name]
        except KeyError:
            if not self.cw_is_saved():
                return None
            rql = "Any A WHERE X eid %%(x)s, X %s A" % name
            try:
                rset = self._cw.execute(rql, {"x": self.eid})
            except Unauthorized:
                self.cw_attr_cache[name] = value = None
            else:
                assert rset.rowcount <= 1, (self, rql, rset.rowcount)
                try:
                    self.cw_attr_cache[name] = value = rset.rows[0][0]
                except IndexError:
                    self.critical(
                        "can't get value for attribute %s of entity with eid %s",
                        name,
                        self.eid,
                    )
                    self.cw_attr_cache[name] = value = None
            return value

    def related(
        self,
        rtype,
        role="subject",
        limit=None,
        entities=False,  # XXX .cw_related
        safe=False,
        targettypes=None,
    ):
        """returns a resultset of related entities

        :param rtype:
          the name of the relation, aka relation type
        :param role:
          the role played by 'self' in the relation ('subject' or 'object')
        :param limit:
          resultset's maximum size
        :param entities:
          if True, the entites are returned; if False, a result set is returned
        :param safe:
          if True, an empty rset/list of entities will be returned in case of
          :exc:`Unauthorized`, else (the default), the exception is propagated
        :param targettypes:
          a tuple of target entity types to restrict the query
        """
        rtype = str(rtype)
        # Caching restricted/limited results is best avoided.
        cacheable = limit is None and targettypes is None
        if cacheable:
            cache_key = f"{rtype}_{role}"
            if cache_key in self._cw_related_cache:
                return self._cw_related_cache[cache_key][entities]
        if not self.has_eid():
            if entities:
                return []
            return self._cw.empty_rset()
        rql = self.cw_related_rql(rtype, role, limit=limit, targettypes=targettypes)
        try:
            rset = self._cw.execute(rql, {"x": self.eid})
        except Unauthorized:
            if not safe:
                raise
            rset = self._cw.empty_rset()
        if cacheable:
            self.cw_set_relation_cache(rtype, role, rset)
        if entities:
            return tuple(rset.entities())
        else:
            return rset

    def cw_related_rql(self, rtype, role="subject", targettypes=None, limit=None):
        return self.cw_related_rqlst(
            rtype, role=role, targettypes=targettypes, limit=limit
        ).as_string()

    def cw_related_rqlst(
        self, rtype, role="subject", targettypes=None, limit=None, sort_terms=None
    ):
        """Return the select node of the RQL query of entities related through
        `rtype` with this entity as `role`, possibly filtered by
        `targettypes`.

        The RQL query can be given a `limit` and sort terms with `sort_terms`
        arguments being a sequence of ``(<relation type>, <sort ascending>)``
        (e.g. ``[('name', True), ('modification_date', False)]`` would lead to
        a sorting by ``name``, ascending and then by ``modification_date``,
        descending. If `sort_terms` is not specified the default sorting is by
        ``modification_date``, descending.
        """
        vreg = self._cw.vreg
        rschema = vreg.schema[rtype]
        select = Select()
        mainvar, evar = select.get_variable("X"), select.get_variable("E")
        select.add_selected(mainvar)
        if limit is not None:
            select.set_limit(limit)
        select.add_eid_restriction(evar, "x", "Substitute")
        if role == "subject":
            rel = make_relation(evar, rtype, (mainvar,), VariableRef)
            select.add_restriction(rel)
            if targettypes is None:
                targettypes = rschema.objects(self.e_schema)
            else:
                select.add_constant_restriction(mainvar, "is", targettypes, "etype")
            gcard = greater_card(rschema, (self.e_schema,), targettypes, 0)
        else:
            rel = make_relation(mainvar, rtype, (evar,), VariableRef)
            select.add_restriction(rel)
            if targettypes is None:
                targettypes = rschema.subjects(self.e_schema)
            else:
                select.add_constant_restriction(mainvar, "is", targettypes, "etype")
            gcard = greater_card(rschema, targettypes, (self.e_schema,), 1)
        etypecls = vreg["etypes"].etype_class(targettypes[0])
        if len(targettypes) > 1:
            fetchattrs = vreg["etypes"].fetch_attrs(targettypes)
            self._fetch_ambiguous_rtypes(
                select, mainvar, fetchattrs, targettypes, vreg.schema
            )
        else:
            fetchattrs = etypecls.fetch_attrs
        etypecls.fetch_rqlst(self._cw.user, select, mainvar, fetchattrs, settype=False)
        # optimisation: remove ORDERBY if cardinality is 1 or ? (though
        # greater_card return 1 for those both cases)
        if gcard == "1":
            select.remove_sort_terms()
        elif not select.orderby:
            # Build a mapping (rtype, node) for relations usable for sorting.
            sorting_relations = {}
            for r in select.where.get_nodes(RqlRelation):
                lhs, rhs = r.children
                if lhs.variable != mainvar:
                    continue
                if r.operator() != "=":
                    continue
                rhs_term = rhs.children[0]
                if not isinstance(rhs_term, VariableRef):
                    continue
                sorting_relations[r.r_type] = r
            sort_terms = sort_terms or [("modification_date", False)]
            for term, order in sort_terms:
                # Add a relation for sorting only if it is not only retrieved
                # (e.g. modification_date) instead of adding another variable
                # for sorting. This should not be problematic.
                rel = sorting_relations.get(term)
                if rel is None:
                    mdvar = select.make_variable()
                    rel = make_relation(mainvar, term, (mdvar,), VariableRef)
                    select.add_restriction(rel)
                var = rel.children[1].children[0].variable
                select.add_sort_var(var, asc=order)
        return select

    # generic vocabulary methods ##############################################

    def cw_linkable_rql(
        self,
        rtype,
        targettype,
        role,
        ordermethod=None,
        vocabconstraints=True,
        lt_infos={},
        limit=None,
    ):
        """build a rql to fetch targettype entities either related or unrelated
        to this entity using (rtype, role) relation.

        Consider relation permissions so that returned entities may be actually
        linked by `rtype`.

        `lt_infos` are supplementary informations, usually coming from __linkto
        parameter, that can help further restricting the results in case current
        entity is not yet created. It is a dict describing entities the current
        entity will be linked to, which keys are (rtype, role) tuples and values
        are a list of eids.
        """
        return self._cw_compute_linkable_rql(
            rtype,
            targettype,
            role,
            ordermethod=None,
            vocabconstraints=vocabconstraints,
            lt_infos=lt_infos,
            limit=limit,
            unrelated_only=False,
        )

    def cw_unrelated_rql(
        self,
        rtype,
        targettype,
        role,
        ordermethod=None,
        vocabconstraints=True,
        lt_infos={},
        limit=None,
    ):
        """build a rql to fetch `targettype` entities unrelated to this entity
        using (rtype, role) relation.

        Consider relation permissions so that returned entities may be actually
        linked by `rtype`.

        `lt_infos` are supplementary informations, usually coming from __linkto
        parameter, that can help further restricting the results in case current
        entity is not yet created. It is a dict describing entities the current
        entity will be linked to, which keys are (rtype, role) tuples and values
        are a list of eids.
        """
        return self._cw_compute_linkable_rql(
            rtype,
            targettype,
            role,
            ordermethod=None,
            vocabconstraints=vocabconstraints,
            lt_infos=lt_infos,
            limit=limit,
            unrelated_only=True,
        )

    def _cw_compute_linkable_rql(
        self,
        rtype,
        targettype,
        role,
        ordermethod=None,
        vocabconstraints=True,
        lt_infos={},
        limit=None,
        unrelated_only=False,
    ):
        """build a rql to fetch `targettype` entities that may be related to
        this entity using the (rtype, role) relation.

        By default (unrelated_only=False), this includes the already linked
        entities as well as the unrelated ones. If `unrelated_only` is True, the
        rql filters out the already related entities.
        """
        ordermethod = ordermethod or "fetch_unrelated_order"
        rschema = self._cw.vreg.schema.relation_schema_for(rtype)
        rdef = rschema.role_relation_definition(self.e_schema, targettype, role)
        rewriter = RQLRewriter(self._cw)
        select = Select()
        # initialize some variables according to the `role` of `self` in the
        # relation (variable names must respect constraints conventions):
        # * variable for myself (`evar`)
        # * variable for searched entities (`searchvedvar`)
        if role == "subject":
            evar = subjvar = select.get_variable("S")
            searchedvar = objvar = select.get_variable("O")
        else:
            searchedvar = subjvar = select.get_variable("S")
            evar = objvar = select.get_variable("O")
        select.add_selected(searchedvar)
        if limit is not None:
            select.set_limit(limit)
        # initialize some variables according to `self` existence
        if rdef.role_cardinality(neg_role(role)) in "?1":
            # if cardinality in '1?', we want a target entity which isn't
            # already linked using this relation
            variable = select.make_variable()
            if role == "subject":
                rel = make_relation(variable, rtype, (searchedvar,), VariableRef)
            else:
                rel = make_relation(searchedvar, rtype, (variable,), VariableRef)
            select.add_restriction(Not(rel))
        elif self.has_eid() and unrelated_only:
            # elif we have an eid, we don't want a target entity which is
            # already linked to ourself through this relation
            rel = make_relation(subjvar, rtype, (objvar,), VariableRef)
            select.add_restriction(Not(rel))
        if self.has_eid():
            rel = make_relation(evar, "eid", ("x", "Substitute"), Constant)
            select.add_restriction(rel)
            args = {"x": self.eid}
            if role == "subject":
                sec_check_args = {"fromeid": self.eid}
            else:
                sec_check_args = {"toeid": self.eid}
            existant = None  # instead of 'SO', improve perfs
        else:
            args = {}
            sec_check_args = {}
            existant = searchedvar.name
            # undefine unused evar, or the type resolver will consider it
            select.undefine_variable(evar)
        # retrieve entity class for targettype to compute base rql
        etypecls = self._cw.vreg["etypes"].etype_class(targettype)
        etypecls.fetch_rqlst(
            self._cw.user, select, searchedvar, ordermethod=ordermethod
        )
        # from now on, we need variable type resolving
        self._cw.vreg.compute_var_types(self._cw, select, args)
        # insert RQL expressions for schema constraints into the rql syntax tree
        if vocabconstraints:
            cstrcls = (RQLVocabularyConstraint, RQLConstraint)
        else:
            cstrcls = RQLConstraint
        lt_infos = pruned_lt_info(self.e_schema, lt_infos or {})
        # if there are still lt_infos, use set to keep track of added eid
        # relations (adding twice the same eid relation is incorrect RQL)
        eidvars = set()
        for cstr in rdef.constraints:
            # consider constraint.mainvars to check if constraint apply
            if isinstance(cstr, cstrcls) and searchedvar.name in cstr.mainvars:
                if not self.has_eid():
                    if lt_infos:
                        # we can perhaps further restrict with linkto infos using
                        # a custom constraint built from cstr and lt_infos
                        cstr = build_cstr_with_linkto_infos(
                            cstr, args, searchedvar, evar, lt_infos, eidvars
                        )
                        if cstr is None:
                            continue  # could not build constraint -> discard
                    elif evar.name in cstr.mainvars:
                        continue
                # compute a varmap suitable to RQLRewriter.rewrite argument
                varmap = dict(
                    (v, v)
                    for v in (searchedvar.name, evar.name)
                    if v in select.defined_vars and v in cstr.mainvars
                )
                # rewrite constraint by constraint since we want a AND between
                # expressions.
                rewriter.rewrite(select, [(varmap, (cstr,))], args, existant)
        # insert security RQL expressions granting the permission to 'add' the
        # relation into the rql syntax tree, if necessary
        rqlexprs = rdef.get_rqlexprs("add")
        if not self.has_eid():
            rqlexprs = [
                rqlexpr for rqlexpr in rqlexprs if searchedvar.name in rqlexpr.mainvars
            ]
        if rqlexprs and not rdef.has_perm(self._cw, "add", **sec_check_args):
            # compute a varmap suitable to RQLRewriter.rewrite argument
            varmap = dict(
                (v, v)
                for v in (searchedvar.name, evar.name)
                if v in select.defined_vars
            )
            # rewrite all expressions at once since we want a OR between them.
            rewriter.rewrite(select, [(varmap, rqlexprs)], args, existant)
        # ensure we have an order defined
        if not select.orderby:
            select.add_sort_var(select.defined_vars[searchedvar.name])
        # we're done, turn the rql syntax tree as a string
        rql = select.as_string()
        return rql, args

    def unrelated(
        self,
        rtype,
        targettype,
        role="subject",
        limit=None,
        ordermethod=None,
        lt_infos={},
    ):  # XXX .cw_unrelated
        """return a result set of target type objects that may be related
        by a given relation, with self as subject or object
        """
        try:
            rql, args = self.cw_unrelated_rql(
                rtype,
                targettype,
                role,
                limit=limit,
                ordermethod=ordermethod,
                lt_infos=lt_infos,
            )
        except Unauthorized:
            return self._cw.empty_rset()
        return self._cw.execute(rql, args)

    # relations cache handling #################################################

    def cw_relation_cached(self, rtype, role):
        """return None if the given relation isn't already cached on the
        instance, else the content of the cache (a 2-uple (rset, entities)).
        """
        return self._cw_related_cache.get(f"{rtype}_{role}")

    def cw_set_relation_cache(self, rtype, role, rset) -> None:
        """set cached values for the given relation"""
        if rset:
            related = tuple(rset.entities(0))
            rschema = self._cw.vreg.schema.relation_schema_for(rtype)
            if role == "subject":
                rcard = rschema.relation_definition(
                    self.e_schema, related[0].e_schema
                ).cardinality[1]
                target = "object"
            else:
                rcard = rschema.relation_definition(
                    related[0].e_schema, self.e_schema
                ).cardinality[0]
                target = "subject"
            if rcard in "?1":
                for rentity in related:
                    rentity._cw_related_cache[f"{rtype}_{target}"] = (
                        self.as_rset(),
                        (self,),
                    )
        else:
            related = ()
        self._cw_related_cache[f"{rtype}_{role}"] = (rset, related)

    def cw_clear_relation_cache(self, rtype=None, role=None) -> None:
        """clear cached values for the given relation or the entire cache if
        no relation is given
        """
        if rtype is None:
            self._cw_related_cache.clear()
            self._cw_adapters_cache.clear()
        else:
            assert role
            self._cw_related_cache.pop(f"{rtype}_{role}", None)

    def cw_clear_all_caches(self) -> None:
        """flush all caches on this entity. Further attributes/relations access
        will triggers new database queries to get back values.

        If you use custom caches on your entity class (take care to @cached!),
        you should override this method to clear them as well.
        """
        # clear attributes cache
        self._cw_completed = False
        self.cw_attr_cache.clear()
        # clear relations cache
        self.cw_clear_relation_cache()
        # rest path unique cache
        try:
            del self.__unique
        except AttributeError:
            pass

    # raw edition utilities ###################################################

    def cw_set(self, **kwargs) -> None:
        """update this entity using given attributes / relation, working in the
        same fashion as :meth:`cw_instantiate`.

        Example (in a shell session):

        >>> c = rql('Any X WHERE X is Company').get_entity(0, 0)
        >>> p = rql('Any X WHERE X is Person').get_entity(0, 0)
        >>> c.cw_set(name=u'Logilab')
        >>> p.cw_set(firstname=u'John', lastname=u'Doe', works_for=c)

        You can also set relations where the entity has 'object' role by
        prefixing the relation name by 'reverse_'.  Also, relation values may be
        an entity or eid, a list of entities or eids, or None (meaning that all
        relations of the given type from or to this object should be deleted).
        """
        assert kwargs
        assert self.cw_is_saved(), (
            "should not call set_attributes while entity " "hasn't been saved yet"
        )
        rql, qargs, pendingrels, attrcache = self._cw_build_entity_query(kwargs)
        if rql:
            rql = "SET " + rql
            qargs["x"] = self.eid
            if " WHERE " in rql:
                rql += ", X eid %(x)s"
            else:
                rql += " WHERE X eid %(x)s"
            self._cw.execute(rql, qargs)
        # update current local object _after_ the rql query to avoid
        # interferences between the query execution itself and the cw_edited /
        # skip_security machinery
        self._cw_update_attr_cache(attrcache)
        self._cw_handle_pending_relations(self.eid, pendingrels, self._cw.execute)
        # XXX update relation cache

    def cw_remove_relation(self, relation_name: str) -> None:
        assert relation_name
        assert isinstance(relation_name, str), type(relation_name)
        self._cw.execute(
            "DELETE X %s Y WHERE X eid %%(x)s" % relation_name, {"x": self.eid}
        )

    def cw_remove_relations(self, relation_names: Iterable[str]) -> None:
        assert relation_names
        assert not isinstance(relation_names, str), type(relation_names)
        for relation_name in relation_names:
            self.cw_remove_relation(relation_name)

    def cw_delete(self, **kwargs) -> None:
        assert self.has_eid(), self.eid
        self._cw.execute(
            "DELETE %s X WHERE X eid %%(x)s" % self.e_schema, {"x": self.eid}, **kwargs
        )

    # server side utilities ####################################################

    def _cw_clear_local_perm_cache(self, action) -> None:
        for rqlexpr in self.e_schema.get_rqlexprs(action):
            self._cw.local_perm_cache.pop((rqlexpr.eid, (("x", self.eid),)), None)


# attribute and relation descriptors ##########################################


class Attribute:
    """descriptor that controls schema attribute access"""

    def __init__(self, attrname):
        assert attrname != "eid"
        self._attrname = attrname

    def __get__(self, eobj, eclass):
        if eobj is None:
            return self
        return eobj.cw_attr_value(self._attrname)


class Relation:
    """descriptor that controls schema relation access"""

    def __init__(self, rschema, role):
        self._rtype = rschema.type
        self._role = role

    def __get__(self, eobj, eclass):
        if eobj is None:
            raise AttributeError(f"{self._rtype} can only be accessed from instances")
        return eobj.related(self._rtype, self._role, entities=True)

    def __set__(self, eobj, value):
        raise NotImplementedError


# entity adapters #############################################################


class Adapter(AppObject):
    """base class for adapters"""

    __registry__ = "adapters"


class EntityAdapter(Adapter):
    """base class for entity adapters (eg adapt an entity to an interface)

    An example would be:

    >>> some_entity.cw_adapt_to('IDownloadable')
    """

    def __init__(self, _cw, **kwargs):
        try:
            self.entity = kwargs.pop("entity")
        except KeyError:
            self.entity = kwargs["rset"].get_entity(
                kwargs.get("row") or 0, kwargs.get("col") or 0
            )
        Adapter.__init__(self, _cw, **kwargs)


set_log_methods(Entity, getLogger("cubicweb.entity"))



# copyright 2014 LOGILAB S.A. (Paris, FRANCE), all rights reserved.
# contact https://www.logilab.fr/ -- mailto:contact@logilab.fr
#
# This file is part of CubicWeb.
#
# CubicWeb is free software: you can redistribute it and/or modify it under the
# terms of the GNU Lesser General Public License as published by the Free
# Software Foundation, either version 2.1 of the License, or (at your option)
# any later version.
#
# CubicWeb is distributed in the hope that it will be useful, but WITHOUT
# ANY WARRANTY; without even the implied warranty of MERCHANTABILITY or FITNESS
# FOR A PARTICULAR PURPOSE.  See the GNU Lesser General Public License for more
# details.
#
# You should have received a copy of the GNU Lesser General Public License along
# with CubicWeb.  If not, see <https://www.gnu.org/licenses/>.
from yams.buildobjs import EntityType, String, Int, SubjectRelation, RelationDefinition

from cubicweb import _

THISYEAR = 2014


class Person(EntityType):
    name = String()
    salaire = Int()
    birth_year = Int(required=True)
    travaille = SubjectRelation("Societe")
    age = Int(formula="Any %d - D WHERE X birth_year D" % THISYEAR)


class Societe(EntityType):
    nom = String()
    salaire_total = Int(
        formula="Any SUM(SA) GROUPBY X WHERE P travaille X, P salaire SA"
    )


class Agent(EntityType):
    asalae_id = String(formula="Any E WHERE M mirror_of X, M extid E")


class MirrorEntity(EntityType):
    extid = String(
        required=True, unique=True, description=_("external identifier of the object")
    )


class mirror_of(RelationDefinition):
    subject = "MirrorEntity"
    object = ("Agent", "Societe")
    cardinality = "?*"
    inlined = True



# copyright 2003-2016 LOGILAB S.A. (Paris, FRANCE), all rights reserved.
# contact https://www.logilab.fr/ -- mailto:contact@logilab.fr
#
# This file is part of CubicWeb.
#
# CubicWeb is free software: you can redistribute it and/or modify it under the
# terms of the GNU Lesser General Public License as published by the Free
# Software Foundation, either version 2.1 of the License, or (at your option)
# any later version.
#
# CubicWeb is distributed in the hope that it will be useful, but WITHOUT
# ANY WARRANTY; without even the implied warranty of MERCHANTABILITY or FITNESS
# FOR A PARTICULAR PURPOSE.  See the GNU Lesser General Public License for more
# details.
#
# You should have received a copy of the GNU Lesser General Public License along
# with CubicWeb.  If not, see <https://www.gnu.org/licenses/>.
"""
.. _ResourceMode:

Resource mode
-------------

Standard resource mode
``````````````````````

A resource *mode* is a predefined set of settings for various resources
directories, such as cubes, instances, etc. to ease development with the
framework. There are two running modes with *CubicWeb*:

* **system**: resources are searched / created in the system directories (eg
  usually requiring root access):

  - instances are stored in :file:`<INSTALL_PREFIX>/etc/cubicweb.d`
  - temporary files (such as pid file) in :file:`<INSTALL_PREFIX>/var/run/cubicweb`

  where `<INSTALL_PREFIX>` is the detected installation prefix ('/usr/local' for
  instance).

* **user**: resources are searched / created in the user home directory:

  - instances are stored in :file:`~/etc/cubicweb.d`
  - temporary files (such as pid file) in :file:`/tmp`


.. _CubicwebWithinVirtualEnv:

Within virtual environment
``````````````````````````

When installed within a virtualenv, CubicWeb will look for instances data as in
user mode by default, that is in $HOME/etc/cubicweb.d. However the
CW_INSTANCES_DIR environment variable should be preferably used.

.. _virtualenv: http://pypi.python.org/pypi/virtualenv


Custom resource location
````````````````````````

Notice that each resource path may be explicitly set using an environment
variable if the default doesn't suit your needs. Here are the default resource
directories that are affected according to mode:

* **system**: ::

        CW_INSTANCES_DIR = <INSTALL_PREFIX>/etc/cubicweb.d/
        CW_INSTANCES_DATA_DIR = <INSTALL_PREFIX>/var/lib/cubicweb/instances/
        CW_RUNTIME_DIR = <INSTALL_PREFIX>/var/run/cubicweb/

* **user**: ::

        CW_INSTANCES_DIR = ~/etc/cubicweb.d/
        CW_INSTANCES_DATA_DIR = ~/etc/cubicweb.d/
        CW_RUNTIME_DIR = /tmp

Cubes search path is also affected, see the :ref:`Cube` section.


Setting Cubicweb Mode
`````````````````````

By default, the mode is set to 'system' for standard installation. The mode is
set to 'user' if `cubicweb is used from a mercurial repository`_. You can force
this by setting the :envvar:`CW_MODE` environment variable to either 'user' or
'system' so you can easily:

* use system wide installation but user specific instances and all, without root
  privileges on the system (`export CW_MODE=user`)

* use local checkout of cubicweb on system wide instances (requires root
  privileges on the system (`export CW_MODE=system`)

If you've a doubt about the mode you're currently running, check the first line
outputed by the :command:`cubicweb-ctl list` command.

.. _`cubicweb is used from a mercurial repository`: CubicwebDevelopmentMod_


.. _CubicwebDevelopmentMod:

Development Mode (source)
`````````````````````````

If :file:`.hg` directory is found into the cubicweb package, there are
specific resource rules.

`<CW_SOFTWARE_ROOT>` is the source checkout's ``cubicweb`` directory:

* cubicweb migration files are searched in `<CW_SOFTWARE_ROOT>/misc/migration`
  instead of `<INSTALL_PREFIX>/share/cubicweb/migration/`.


Development Mode (virtualenv)
`````````````````````````````

If a virtualenv is found to be activated (i.e. a VIRTUAL_ENV variable is found
in environment), the virtualenv root is used as `<INSTALL_PREFIX>`. This, in
particular, makes it possible to work in `setuptools development mode`_
(``python setup.py develop``) without any further configuration.

.. _`setuptools development mode`: https://pythonhosted.org/setuptools/setuptools.html#development-mode

.. _ConfigurationEnv:

Environment configuration
-------------------------

Python
``````

If you installed *CubicWeb* by cloning the Mercurial shell repository or from source
distribution, then you will need to update the environment variable PYTHONPATH by
adding the path to `cubicweb`:

Add the following lines to either :file:`.bashrc` or :file:`.bash_profile` to
configure your development environment ::

    export PYTHONPATH=/full/path/to/grshell-cubicweb

If you installed *CubicWeb* with packages, no configuration is required and your
new cubes will be placed in `/usr/share/cubicweb/cubes` and your instances will
be placed in `/etc/cubicweb.d`.


CubicWeb
````````

Here are all environment variables that may be used to configure *CubicWeb*:

.. envvar:: CW_MODE

   Resource mode: user or system, as explained in :ref:`ResourceMode`.

.. envvar:: CW_INSTANCES_DIR

   Directory where cubicweb instances will be found.

.. envvar:: CW_INSTANCES_DATA_DIR

   Directory where cubicweb instances data will be written (backup file...)

.. envvar:: CW_RUNTIME_DIR

   Directory where pid files will be written
"""

import importlib
import logging
import logging.config
import os
import pkgutil
import stat
import sys
from os.path import (
    exists,
    join,
    expanduser,
    abspath,
    basename,
    dirname,
    splitext,
    realpath,
)
from smtplib import SMTP, SMTPAuthenticationError
from threading import Lock
from warnings import filterwarnings

from logilab.common.configuration import (
    Configuration,
    Method,
    ConfigurationMixIn,
    merge_options,
    _validate as lgc_validate,
    REQUIRED,
    OptionError,
)
from logilab.common.decorators import cached
from logilab.common.logging_ext import set_log_methods, init_log

from cubicweb import CW_MIGRATION_MAP, ConfigurationError, Binary, _
from cubicweb.server import SOURCE_TYPES
from cubicweb.server.serverconfig import SourceConfiguration, generate_source_config
from cubicweb.toolsutils import (
    create_dir,
    option_value_from_env,
    read_config,
    restrict_perms_to_user,
)

if sys.version_info < (3, 10):
    import importlib_metadata
else:
    import importlib.metadata as importlib_metadata

CONFIG_HELP_MESSAGE = """# This file is generated by CubicWeb. It describes all instance options.
    # All these options can be overridden by environment variables, which should
    # be named with the following pattern: CW_<OPTION_NAME> where OPTION_NAME is
    # the name of the option in uppercase and '-' converted in '_'.
    #
    # For example, the host option is replaced by the environment variable CW_HOST.
    # connections-pooler-enabled is overridden by CW_CONNECTIONS_POOLER_ENABLED

    """

SMTP_LOCK = Lock()


def _cube_pkgname(cube):
    if not cube.startswith("cubicweb_"):
        return "cubicweb_" + cube
    return cube


def _expand_modname(modname, recursive=True):
    """expand modules names `modname` if exists by recursively walking
    submodules and subpackages and yield (submodname, filepath) including
    `modname` itself

    If the file ends with .pyc or .pyo (python bytecode) also check that the
    corresponding source .py file exists before yielding.

    If `recursive` is False skip subpackages.
    """
    try:
        spec = importlib.util.find_spec(modname)
    except ImportError:
        return
    if not spec:
        return

    def check_source_file(filepath):
        if filepath[-4:] in (".pyc", ".pyo"):
            if not exists(filepath[:-1]):
                return False
        return True

    filepath = spec.origin
    if filepath is None or not check_source_file(filepath):
        return
    yield modname, filepath
    loader = spec.loader
    if loader.is_package(modname):
        path = dirname(filepath)
        for subloader, subname, ispkg in pkgutil.walk_packages([path]):
            submodname = ".".join([modname, subname])
            if not ispkg:
                filepath = subloader.find_spec(subname).origin
                if check_source_file(filepath):
                    yield submodname, filepath
            elif recursive:
                for x in _expand_modname(submodname, recursive=True):
                    yield x


# persistent options definition
PERSISTENT_OPTIONS = (
    (
        "encoding",
        {
            "type": "string",
            "default": "UTF-8",
            "help": _("user interface encoding"),
            "group": "ui",
            "sitewide": True,
        },
    ),
    (
        "language",
        {
            "type": "string",
            "default": "en",
            "vocabulary": Method("available_languages"),
            "help": _("language of the user interface"),
            "group": "ui",
        },
    ),
    (
        "date-format",
        {
            "type": "string",
            "default": "%Y/%m/%d",
            "help": _(
                'how to format date in the ui (see <a href="http://docs.python.org/library/datetime.html#strftime-strptime-behavior">this page</a> for format description)'
            ),
            "group": "ui",
        },
    ),
    (
        "datetime-format",
        {
            "type": "string",
            "default": "%Y/%m/%d %H:%M",
            "help": _(
                'how to format date and time in the ui (see <a href="http://docs.python.org/library/datetime.html#strftime-strptime-behavior">this page</a> for format description)'
            ),
            "group": "ui",
        },
    ),
    (
        "time-format",
        {
            "type": "string",
            "default": "%H:%M",
            "help": _(
                'how to format time in the ui (see <a href="http://docs.python.org/library/datetime.html#strftime-strptime-behavior">this page</a> for format description)'
            ),
            "group": "ui",
        },
    ),
    (
        "float-format",
        {
            "type": "string",
            "default": "%.3f",
            "help": _("how to format float numbers in the ui"),
            "group": "ui",
        },
    ),
    (
        "default-text-format",
        {
            "type": "choice",
            "choices": ("text/plain", "text/rest", "text/html", "text/markdown"),
            "default": "text/plain",
            "help": _("default text format for rich text fields."),
            "group": "ui",
        },
    ),
    (
        "short-line-size",
        {
            "type": "int",
            "default": 80,
            "help": _("maximum number of characters in short description"),
            "group": "navigation",
        },
    ),
)


def register_persistent_options(options):
    global PERSISTENT_OPTIONS
    PERSISTENT_OPTIONS = merge_options(PERSISTENT_OPTIONS + options)


CFGTYPE2ETYPE_MAP = {
    "string": "String",
    "choice": "String",
    "yn": "Boolean",
    "int": "Int",
    "float": "Float",
}


_INSTALL_PREFIX = os.environ.get("CW_INSTALL_PREFIX", sys.prefix)
_USR_INSTALL = _INSTALL_PREFIX == "/usr"


class CubicWebConfiguration(ConfigurationMixIn):
    """base class for cubicweb server and web configurations"""

    # to set in concrete configuration
    name = "repository"
    # log messages format (see logging module documentation for available keys)
    log_format = "%(asctime)s - (%(name)s) %(levelname)s: %(message)s"
    # the format below can be useful to debug multi thread issues:
    # log_format = '%(asctime)s - [%(threadName)s] (%(name)s) %(levelname)s: %(message)s'
    # nor remove appobjects based on unused interface [???]
    cleanup_unused_appobjects = True

    # read the schema from the database
    read_instance_schema = True
    # set this to true to get a minimal repository, for instance to get cubes
    # information on commands such as i18ninstance, db-restore, etc...
    quick_start = False
    # check user's state at login time
    consider_user_state = True

    # should some hooks be deactivated during [pre|post]create script execution
    free_wheel = False

    # list of enables sources when sources restriction is necessary
    # (eg repository initialization at least)
    enabled_sources = None

    if "VIRTUAL_ENV" in os.environ:
        mode = os.environ.get("CW_MODE", "user")
    else:
        mode = os.environ.get("CW_MODE", "system")
    assert mode in ("system", "user"), '"CW_MODE" should be either "user" or "system"'

    if mode == "user":
        _INSTANCES_DIR = expanduser("~/etc/cubicweb.d/")
    # mode == system'
    elif _USR_INSTALL:
        _INSTANCES_DIR = "/etc/cubicweb.d/"
    else:
        _INSTANCES_DIR = join(_INSTALL_PREFIX, "etc", "cubicweb.d")

    # set to true during repair (shell, migration) to allow some things which
    # wouldn't be possible otherwise
    repairing = False

    # set by upgrade command
    verbosity = 0
    cmdline_options = None

    smtp_cls = SMTP

    options = (
        (
            "log-threshold",
            {
                "type": "string",  # XXX use a dedicated type?
                "default": "WARNING",
                "help": "server's log level",
                "group": "main",
                "level": 1,
            },
        ),
        (
            "umask",
            {
                "type": "int",
                "default": 0o077,
                "help": "permission umask for files created by the server",
                "group": "main",
                "level": 2,
            },
        ),
        # common configuration options which are potentially required as soon as
        # you're using "base" application objects (ie to really server/web
        # specific)
        (
            "base-url",
            {
                "type": "string",
                "default": None,
                "help": "web server root url",
                "group": "main",
                "level": 1,
            },
        ),
        (
            "receives-base-url-path",
            {
                "type": "yn",
                "default": True,
                "help": "will the instance receives and handle a sub path if present on its base-url",
                "group": "main",
                "level": 1,
            },
        ),
        (
            "allow-email-login",
            {
                "type": "yn",
                "default": False,
                "help": "allow users to login with their primary email if set",
                "group": "main",
                "level": 2,
            },
        ),
        (
            "mangle-emails",
            {
                "type": "yn",
                "default": False,
                "help": "don't display actual email addresses but mangle them if \
this option is set to yes",
                "group": "email",
                "level": 3,
            },
        ),
        (
            "debug",
            {
                "type": "yn",
                "default": False,
                "help": "activate debug mode on startup, disabled by default",
                "group": "main",
                "level": 3,
            },
        ),
        (
            "log-file",
            {
                "type": "string",
                "default": Method("default_log_file"),
                "help": "file where output logs should be written",
                "group": "main",
                "level": 2,
            },
        ),
        # email configuration
        (
            "smtp-host",
            {
                "type": "string",
                "default": "mail",
                "help": "hostname of the SMTP mail server",
                "group": "email",
                "level": 1,
            },
        ),
        (
            "smtp-port",
            {
                "type": "int",
                "default": 25,
                "help": "listening port of the SMTP mail server",
                "group": "email",
                "level": 1,
            },
        ),
        (
            "smtp-username",
            {
                "default": "",
                "help": "username for SMTP mail server if auth is required",
                "group": "email",
                "level": 2,
            },
        ),
        (
            "smtp-password",
            {
                "default": "",
                "help": "password for SMTP mail server if auth is required",
                "group": "email",
                "level": 2,
            },
        ),
        (
            "sender-name",
            {
                "type": "string",
                "default": Method("default_instance_id"),
                "help": (
                    "name used as HELO name for outgoing emails from the repository."
                ),
                "group": "email",
                "level": 2,
            },
        ),
        (
            "sender-addr",
            {
                "type": "string",
                "default": "cubicweb@mydomain.com",
                "help": (
                    "email address used as HELO address for outgoing emails from the "
                    "repository"
                ),
                "group": "email",
                "level": 1,
            },
        ),
        (
            "sender-x-cw-header",
            {
                "type": "string",
                "default": Method("default_instance_id"),
                "help": "Value of the header X-CW for outgoing emails.",
                "group": "email",
                "level": 3,
            },
        ),
        (
            "logstat-interval",
            {
                "type": "int",
                "default": 0,
                "help": (
                    "interval (in seconds) at which stats are dumped in the logstat file; set 0 to "
                    "disable"
                ),
                "group": "main",
                "level": 2,
            },
        ),
        (
            "logstat-file",
            {
                "type": "string",
                "default": Method("default_stats_file"),
                "help": "file where stats for the instance should be written",
                "group": "main",
                "level": 2,
            },
        ),
        # ctl configuration
        (
            "host",
            {
                "type": "string",
                "default": "localhost",
                "help": "host name if not correctly detectable through gethostname",
                "group": "main",
                "level": 1,
            },
        ),
        (
            "pid-file",
            {
                "type": "string",
                "default": Method("default_pid_file"),
                "help": "repository's pid file",
                "group": "main",
                "level": 2,
            },
        ),
        (
            "uid",
            {
                "type": "string",
                "default": None,
                "help": (
                    "unix user, if this option is set, use the specified user to start \
the repository rather than the user running the command"
                ),
                "group": "main",
                "level": (mode == "installed") and 0 or 1,
            },
        ),
        (
            "cleanup-session-time",
            {
                "type": "time",
                "default": "24h",
                "help": (
                    "duration of inactivity after which a session "
                    "will be closed, to limit memory consumption (avoid sessions that "
                    "never expire and cause memory leak when http-session-time is 0, or "
                    "because of bad client that never closes their connection). "
                    "So notice that even if http-session-time is 0 and the user don't "
                    "close his browser, he will have to reauthenticate after this time "
                    "of inactivity. Default to 24h."
                ),
                "group": "main",
                "level": 3,
            },
        ),
        (
            "connections-pooler-enabled",
            {
                "type": "yn",
                "default": True,
                "help": "Enable the connection pooler. Set to no if you use an external database pooler (e.g. pgbouncer)",
                "group": "main",
                "level": 3,
            },
        ),
        (
            "connections-pool-max-size",
            {
                "type": "int",
                "default": 0,
                "help": "Maximum, per process, number of database connections. Default 0 (unlimited)",
                "group": "main",
                "level": 3,
            },
        ),
        (
            "connections-pool-min-size",
            {
                "type": "int",
                "default": 0,
                "help": "Minimum, per process, number of database connections.",
                "group": "main",
                "level": 3,
            },
        ),
        (
            "connections-pool-idle-timeout",
            {
                "type": "int",
                "default": 600,
                "help": "Start closing connection if the pool hasn't been empty for this many seconds",
                "group": "main",
                "level": 3,
            },
        ),
        (
            "rql-cache-size",
            {
                "type": "int",
                "default": 3000,
                "help": "size of the parsed rql cache size.",
                "group": "main",
                "level": 3,
            },
        ),
        (
            "undo-enabled",
            {
                "type": "yn",
                "default": False,
                "help": "enable undo support",
                "group": "main",
                "level": 3,
            },
        ),
        (
            "keep-transaction-lifetime",
            {
                "type": "int",
                "default": 7,
                "help": "number of days during which transaction records should be \
kept (hence undoable).",
                "group": "main",
                "level": 3,
            },
        ),
        (
            "multi-sources-etypes",
            {
                "type": "csv",
                "default": (),
                "help": "defines which entity types from this repository are used \
by some other instances. You should set this properly for these instances to \
detect updates / deletions.",
                "group": "main",
                "level": 3,
            },
        ),
        (
            "delay-full-text-indexation",
            {
                "type": "yn",
                "default": False,
                "help": (
                    "When full text indexation of entity has a too important cost"
                    " to be done when entity are added/modified by users, activate this "
                    "option and setup a job using cubicweb-ctl db-rebuild-fti on your "
                    "system (using cron for instance)."
                ),
                "group": "main",
                "level": 3,
            },
        ),
        # email configuration
        (
            "default-recipients-mode",
            {
                "type": "choice",
                "choices": ("default-dest-addrs", "users", "none"),
                "default": "default-dest-addrs",
                "help": 'when a notification should be sent with no specific rules \
to find recipients, recipients will be found according to this mode. Available \
modes are "default-dest-addrs" (emails specified in the configuration \
variable with the same name), "users" (every users which has activated \
account with an email set), "none" (no notification).',
                "group": "email",
                "level": 2,
            },
        ),
        (
            "default-dest-addrs",
            {
                "type": "csv",
                "default": (),
                "help": "comma separated list of email addresses that will be used \
as default recipient when an email is sent and the notification has no \
specific recipient rules.",
                "group": "email",
                "level": 2,
            },
        ),
        (
            "supervising-addrs",
            {
                "type": "csv",
                "default": (),
                "help": "comma separated list of email addresses that will be \
notified of every changes.",
                "group": "email",
                "level": 2,
            },
        ),
        (
            "anonymous-user",
            {
                "type": "string",
                "default": None,
                "help": (
                    "login of the CubicWeb user account to use for anonymous "
                    "user (if you want to allow anonymous)"
                ),
                "group": "web",
                "level": 1,
            },
        ),
        (
            "anonymous-password",
            {
                "type": "string",
                "default": None,
                "help": (
                    "password of the CubicWeb user account to use for anonymous user, "
                    "if anonymous-user is set"
                ),
                "group": "web",
                "level": 1,
            },
        ),
        (
            "cleanup-anonymous-session-time",
            {
                "type": "time",
                "default": "5min",
                "help": (
                    "Same as cleanup-session-time but specific to anonymous "
                    "sessions. You can have a much smaller timeout here since it will be "
                    "transparent to the user. Default to 5min."
                ),
                "group": "web",
                "level": 3,
            },
        ),
        # dbdump configuration
        (
            "dump-s3-endpoint-url",
            {
                "type": "string",
                "default": None,
                "help": "S3 endpoint url for database dump/restore (without protocol)",
                "group": "dbdump",
                "level": 3,
            },
        ),
        (
            "dump-s3-bucket-name",
            {
                "type": "string",
                "default": None,
                "help": "S3 bucket name for database dump/restore",
                "group": "dbdump",
                "level": 3,
            },
        ),
        (
            "dump-s3-access-key",
            {
                "type": "string",
                "default": None,
                "help": "S3 access key for database dump/restore",
                "group": "dbdump",
                "level": 3,
            },
        ),
        (
            "dump-s3-secret-key",
            {
                "type": "string",
                "default": None,
                "help": "S3 secret key for database dump/restore",
                "group": "dbdump",
                "level": 3,
            },
        ),
        (
            "dump-s3-region",
            {
                "type": "string",
                "default": None,
                "help": "S3 region name of buckets for database dump/restore",
                "group": "dbdump",
                "level": 3,
            },
        ),
        (
            "dump-s3-secure",
            {
                "type": "yn",
                "default": True,
                "help": "Activate the secure flag to use TLS connection with S3",
                "group": "dbdump",
                "level": 3,
            },
        ),
        (
            "dump-s3-cert-check",
            {
                "type": "yn",
                "default": True,
                "help": "Activate the verification of the certificat for HTTPS connections",
                "group": "dbdump",
                "level": 3,
            },
        ),
    )

    def load_defaults(self) -> None:
        """overload the parent `load_defaults` to load REQUIRED variables with
        environment values
        """
        for opt, optdict in self.options:
            action = optdict.get("action")
            if action != "callback":
                # callback action have no default
                default = self.option_default(opt, optdict)
                if default is REQUIRED:
                    # the next two lines are different from the parent's method.
                    # In the parent's, we always continue.
                    # In this case, we continue only if the value can not be
                    # read from the environment
                    default = option_value_from_env(opt)
                    if default is None:
                        continue
                self.set_option(opt, default, action, optdict)

    def __getitem__(self, key):
        """Get configuration option, by first looking at environment."""
        file_value = super(CubicWebConfiguration, self).__getitem__(key)
        value = option_value_from_env(key, file_value)
        if value is not None:
            option_def = self.get_option_def(key)
            value = lgc_validate(value, option_def)
        return value

    # static and class methods used to get instance independant resources ##
    @staticmethod
    def cubicweb_version():
        """return installed cubicweb version"""
        from logilab.common.changelog import Version

        str_base_version = importlib_metadata.version("cubicweb")
        version = tuple([int(x) for x in str_base_version.split(".")])
        assert len(version) == 3, version
        return Version(version)

    @staticmethod
    def persistent_options_configuration():
        return Configuration(options=PERSISTENT_OPTIONS)

    @classmethod
    def i18n_lib_dir(cls):
        """return instance's i18n directory"""
        return join(dirname(__file__), "i18n")

    @classmethod
    def cw_languages(cls):
        for fname in os.listdir(join(cls.i18n_lib_dir())):
            if fname.endswith(".po"):
                yield splitext(fname)[0]

    @classmethod
    def available_cubes(cls):
        """Return a list of available cube names.

        For cube as package, name is equal to python package's name.
        """
        cubes = set()
        for entry_point in importlib_metadata.entry_points(group="cubicweb.cubes"):
            try:
                module = entry_point.load()
            except ImportError:
                continue
            else:
                modname = module.__name__
                if not modname.startswith("cubicweb_"):
                    cls.warning(
                        "entry point %s does not appear to be a cube", entry_point
                    )
                    continue
                cubes.add(modname)

        def sortkey(cube):
            """Preserve sorting with "cubicweb_" prefix."""
            prefix = "cubicweb_"
            if cube.startswith(prefix):
                # add a suffix to have a deterministic sorting between
                # 'cubicweb_<cube>' and '<cube>' (useful in tests with "hash
                # randomization" turned on).
                return cube[len(prefix) :] + "~"
            return cube

        return sorted(cubes, key=sortkey)

    @classmethod
    def cube_dir(cls, cube):
        """return the cube directory for the given cube id, raise
        `ConfigurationError` if it doesn't exist
        """
        pkgname = _cube_pkgname(cube)
        loader = importlib.util.find_spec(pkgname)
        if loader:
            return dirname(loader.origin)
        raise ConfigurationError(
            f"no module {pkgname} while searching in cube '{cube}'"
        )

    @classmethod
    def cube_migration_scripts_dir(cls, cube):
        """cube migration scripts directory"""
        return join(cls.cube_dir(cube), "migration")

    @classmethod
    def cube_pkginfo(cls, cube):
        """return the information module for the given cube"""
        cube = CW_MIGRATION_MAP.get(cube, cube)
        pkgname = _cube_pkgname(cube)
        return importlib.import_module(f"{pkgname}.__pkginfo__")

    @classmethod
    def cube_version(cls, cube):
        """return the version of the cube located in the given directory"""
        from logilab.common.changelog import Version

        version = cls.cube_pkginfo(cube).numversion
        assert len(version) == 3, version
        return Version(version)

    @classmethod
    def _cube_deps(cls, cube, key):
        """return cubicweb cubes used by the given cube"""
        pkginfo = cls.cube_pkginfo(cube)
        try:
            # explicit __xxx_cubes__ attribute
            deps = getattr(pkginfo, key)
        except AttributeError:
            # deduce cubes from generic __xxx__ attribute
            try:
                gendeps = getattr(pkginfo, key.replace("_cubes", ""))
            except AttributeError:
                deps = {}
            else:
                deps = dict(
                    (x[len("cubicweb-") :], v)
                    for x, v in gendeps.items()
                    if x.startswith("cubicweb-")
                )
        for depcube in deps:
            try:
                newname = CW_MIGRATION_MAP[depcube]
            except KeyError:
                pass
            else:
                deps[newname] = deps.pop(depcube)
        return deps

    @classmethod
    def cube_depends_cubicweb_version(cls, cube):
        # XXX no backward compat (see _cube_deps above)
        try:
            pkginfo = cls.cube_pkginfo(cube)
            deps = getattr(pkginfo, "__depends__")
            return deps.get("cubicweb")
        except AttributeError:
            return None

    @classmethod
    def cube_dependencies(cls, cube):
        """return cubicweb cubes used by the given cube"""
        return cls._cube_deps(cube, "__depends_cubes__")

    @classmethod
    def cube_recommends(cls, cube):
        """return cubicweb cubes recommended by the given cube"""
        return cls._cube_deps(cube, "__recommends_cubes__")

    @classmethod
    def expand_cubes(cls, cubes, with_recommends=False):
        """expand the given list of top level cubes used by adding recursivly
        each cube dependencies
        """
        cubes = list(cubes)
        todo = cubes[:]
        if with_recommends:
            available = set(cls.available_cubes())
        while todo:
            cube = todo.pop(0)
            for depcube in cls.cube_dependencies(cube):
                if depcube not in cubes:
                    cubes.append(depcube)
                    todo.append(depcube)
            if with_recommends:
                for depcube in cls.cube_recommends(cube):
                    if depcube not in cubes and depcube in available:
                        cubes.append(depcube)
                        todo.append(depcube)
        return cubes

    @classmethod
    def reorder_cubes(cls, cubes):
        """reorder cubes from the top level cubes to inner dependencies
        cubes
        """
        from logilab.common.graph import ordered_nodes, UnorderableGraph

        graph = {}
        for cube in cubes:
            cube = CW_MIGRATION_MAP.get(cube, cube)
            graph[cube] = set(
                dep for dep in cls.cube_dependencies(cube) if dep in cubes
            )
            graph[cube] |= set(dep for dep in cls.cube_recommends(cube) if dep in cubes)
        try:
            return ordered_nodes(graph)
        except UnorderableGraph as ex:
            raise ConfigurationError(ex)

    @classmethod
    def load_available_configs(cls):
        for confmod in (
            "cubicweb_web.webconfig",
            "cubicweb.server.serverconfig",
            "cubicweb.pyramid.config",
        ):
            try:
                __import__(confmod)
            except ImportError as exc:
                cls.warning("failed to load config module %s (%s)", confmod, exc)

    @classmethod
    def load_cwctl_plugins(cls):
        for ctlmod in (
            "cubicweb.server.serverctl",
            "cubicweb.devtools.devctl",
            "cubicweb.pyramid.pyramidctl",
        ):
            try:
                __import__(ctlmod)
            except ImportError as exc:
                cls.warning("failed to load cubicweb-ctl plugin %s (%s)", ctlmod, exc)
                continue
            cls.info("loaded cubicweb-ctl plugin %s", ctlmod)

        for cube in cls.available_cubes():
            cubedir = cls.cube_dir(cube)
            pluginfile = join(cubedir, "ccplugin.py")
            initfile = join(cubedir, "__init__.py")
            pkgname = _cube_pkgname(cube)
            if exists(pluginfile):
                try:
                    __import__(pkgname + ".ccplugin")
                    cls.info("loaded cubicweb-ctl plugin from %s", cube)
                except Exception:
                    cls.exception("while loading plugin %s", pluginfile)
            elif exists(initfile):
                try:
                    __import__(pkgname)
                except Exception:
                    cls.exception("while loading cube %s", cube)
            else:
                cls.warning("no __init__ file in cube %s", cube)

    cubicweb_appobject_path = {"entities", "sobjects", "hooks"}
    cube_appobject_path = {"entities", "sobjects", "hooks"}

    def __init__(self, appid, debugmode=False, creating=False, log_to_file=False):
        self.appid = appid

        # set to true while creating an instance
        self.creating = creating

        if debugmode:
            # in python 2.7, DeprecationWarning are not shown anymore by default
            filterwarnings("default", category=DeprecationWarning)

        register_stored_procedures()
        self._cubes = None
        super(CubicWebConfiguration, self).__init__()
        self.debugmode = debugmode
        self.log_to_file = log_to_file
        self.adjust_sys_path()
        self.load_defaults()

        # will be properly initialized later by _gettext_init
        self.translations = {"en": (str, lambda ctx, msgid: str(msgid))}
        self._site_loaded = set()

        fake_gettext = (str, lambda ctx, msgid: str(msgid))
        for lang in self.available_languages():
            self.translations[lang] = fake_gettext

        self._cubes = None
        self.load_file_configuration(self.main_config_file())

    def adjust_sys_path(self):
        # adding apphome to python path is not usually necessary in production
        # environments, but necessary for tests
        if self.apphome and self.apphome not in sys.path:
            sys.path.insert(0, self.apphome)

    def init_log(self, logthreshold=None, force=False, logfile=None, syslog=None):
        """init the log service"""
        if not force and hasattr(self, "_logging_initialized"):
            return
        self._logging_initialized = True

        if logthreshold is None:
            if self.debugmode:
                logthreshold = "DEBUG"
            else:
                logthreshold = self["log-threshold"]
        init_log(not self.log_to_file, syslog, logthreshold, logfile, self.log_format)

        # configure simpleTal logger
        logging.getLogger("simpleTAL").setLevel(logging.ERROR)

        self.init_log(logthreshold, logfile=self.get("log-file"))
        # read a config file if it exists
        logconfig = join(self.apphome, "logging.conf")
        if exists(logconfig):
            logging.config.fileConfig(logconfig)

    def schema_modnames(self):
        modnames = []
        for name in ("bootstrap", "base", "workflow", "Bookmark"):
            modnames.append(("cubicweb", "cubicweb.schemas." + name))
        for cube in reversed(self.cubes()):
            for modname, filepath in _expand_modname(
                f"{_cube_pkgname(cube)}.schema", recursive=False
            ):
                modnames.append((cube, modname))
        if self.apphome:
            apphome = realpath(self.apphome)
            for modname, filepath in _expand_modname("schema", recursive=False):
                if realpath(filepath).startswith(apphome):
                    modnames.append(("data", modname))
        return modnames

    def appobjects_modnames(self):
        modnames = []
        for name in self._sorted_appobjects(self.cubicweb_appobject_path):
            for modname, filepath in _expand_modname("cubicweb." + name):
                modnames.append(modname)

        for cube in reversed(self.cubes()):
            modnames.extend(self.appobjects_cube_modnames(cube))

        if self.apphome:
            cube_submodnames = self._sorted_appobjects(self.cube_appobject_path)
            apphome = realpath(self.apphome)
            for name in cube_submodnames:
                for modname, filepath in _expand_modname(name):
                    # ensure file is in apphome
                    if realpath(filepath).startswith(apphome):
                        modnames.append(modname)

        return modnames

    @property
    def apphome(self):
        return join(self.instances_dir(), self.appid)

    def load_site_cubicweb(self, cubes=()):
        """load site_cubicweb file for `cubes`"""
        for cube in reversed(cubes or self.cubes()):
            if cube in self._site_loaded:
                continue
            try:
                self._load_site_cubicweb(cube)
                self._site_loaded.add(cube)
            except ImportError:
                continue
        if self.apphome is not None:
            # Would occur, e.g., upon `cubicweb-ctl i18ncube <cube>`.
            self._load_site_cubicweb(None)

    def _load_site_cubicweb(self, cube):
        """Load site_cubicweb.py from `cube` (or apphome if cube is None)."""
        mod = None

        if cube is not None:
            modname = _cube_pkgname(cube)
            __import__(modname)
            modname = modname + ".site_cubicweb"
            __import__(modname)
            mod = sys.modules[modname]
        else:
            import types
            import importlib.machinery

            apphome_site = join(self.apphome, "site_cubicweb.py")
            if exists(apphome_site):
                loader = importlib.machinery.SourceFileLoader(
                    "site_cubicweb", apphome_site
                )
                mod = types.ModuleType(loader.name)
                loader.exec_module(mod)
            else:
                mod = None

        # overridden to register cube specific options
        if getattr(mod, "options", None):
            self.register_options(mod.options)
            self.load_defaults()

    def cwproperty_definitions(self):
        cfg = self.persistent_options_configuration()
        for section, options in cfg.options_by_section():
            section = section.lower()
            for optname, optdict, value in options:
                key = f"{section}.{optname}"
                type, vocab = self.map_option(optdict)
                default = cfg.option_default(optname, optdict)
                pdef = {
                    "type": type,
                    "vocabulary": vocab,
                    "default": default,
                    "help": optdict["help"],
                    "sitewide": optdict.get("sitewide", False),
                }
                yield key, pdef

    def map_option(self, optdict):
        try:
            vocab = optdict["choices"]
        except KeyError:
            vocab = optdict.get("vocabulary")
            if isinstance(vocab, Method):
                vocab = getattr(self, vocab.method, ())
        return CFGTYPE2ETYPE_MAP[optdict["type"]], vocab

    def default_instance_id(self):
        """return the instance identifier, useful for option which need this
        as default value
        """
        return self.appid

    _cubes = None

    @classmethod
    def _warn_pyramid_cube(cls):
        cls.warning(
            "cubicweb-pyramid got integrated into CubicWeb; "
            "remove it from your project's dependencies"
        )

    def init_cubes(self, cubes):
        cubes = list(cubes)
        if "pyramid" in cubes:
            self._warn_pyramid_cube()
            cubes.remove("pyramid")

        self._cubes = self.reorder_cubes(cubes)

        # load cubes'__init__.py file first
        for cube in cubes:
            importlib.import_module(_cube_pkgname(cube))

        self.load_site_cubicweb()

        # reload config file in cases options are defined in cubes __init__
        # or site_cubicweb files
        self.load_file_configuration(self.main_config_file())

        # configuration initialization hook
        self.load_configuration(**(self.cmdline_options or {}))

    def cubes(self):
        """return the list of cubes used by this instance

        result is ordered from the top level cubes to inner dependencies
        cubes
        """
        assert self._cubes is not None, "cubes not initialized"
        return self._cubes

    def cubes_path(self):
        """return the list of path to cubes used by this instance, from outer
        most to inner most cubes
        """
        return [self.cube_dir(p) for p in self.cubes()]

    # these are overridden by set_log_methods below
    # only defining here to prevent pylint from complaining
    @classmethod
    def debug(cls, msg, *a, **kw):
        pass

    @classmethod
    def instances_dir(cls):
        """return the control directory"""
        return abspath(os.environ.get("CW_INSTANCES_DIR", cls._INSTANCES_DIR))

    @classmethod
    def migration_scripts_dir(cls):
        """cubicweb migration scripts directory"""
        mdir = join(dirname(__file__), "misc", "migration")
        assert exists(mdir), f"migration path {mdir} does not exist"
        return mdir

    @classmethod
    def config_for(
        cls, appid, config=None, debugmode=False, log_to_file=False, creating=False
    ):
        """return a configuration instance for the given instance identifier"""
        cls.load_available_configs()
        try:
            from cubicweb_web.webconfig import WebAllInOneConfiguration

            configcls = WebAllInOneConfiguration
        except ImportError:
            from cubicweb.pyramid.config import AllInOneConfiguration

            configcls = AllInOneConfiguration

        return configcls(appid, debugmode, creating, log_to_file=log_to_file)

    @classmethod
    def instance_home(cls, appid):
        """return the home directory of the instance with the given
        instance id
        """
        home = join(cls.instances_dir(), appid)
        if not exists(home):
            raise ConfigurationError(
                f'no such instance {appid} (check it exists with "cubicweb-ctl list")'
            )
        return home

    MODES = ("common", "repository", "Any")
    MCOMPAT = {
        "all-in-one": MODES,
    }

    @classmethod
    def accept_mode(cls, mode):
        # assert mode in cls.MODES, mode
        return mode in cls.MCOMPAT[cls.name]

    # default configuration methods ###########################################

    def default_log_file(self):
        """return default path to the log file of the instance'server"""
        if self.mode == "user":
            import tempfile

            basepath = join(
                tempfile.gettempdir(), f"{basename(self.appid)}-{self.name}"
            )
            path = basepath + ".log"
            i = 1
            while exists(path) and i < 100:  # arbitrary limit to avoid infinite loop
                try:
                    open(path, "a")
                    break
                except IOError:
                    path = f"{basepath}-{i}.log"
                    i += 1
            return path
        if _USR_INSTALL:
            return f"/var/log/cubicweb/{self.appid}-{self.name}.log"
        else:
            log_path = os.path.join(
                _INSTALL_PREFIX, "var", "log", "cubicweb", "%s-%s.log"
            )
            return log_path % (self.appid, self.name)

    def default_stats_file(self):
        """return default path to the stats file of the instance'server"""
        logfile = self.default_log_file()
        if logfile.endswith(".log"):
            logfile = logfile[:-4]
        return logfile + ".stats"

    def default_pid_file(self):
        """return default path to the pid file of the instance'server"""
        if self.mode == "system":
            if _USR_INSTALL:
                default = "/var/run/cubicweb/"
            else:
                default = os.path.join(_INSTALL_PREFIX, "var", "run", "cubicweb")
        else:
            import tempfile

            default = tempfile.gettempdir()
        # runtime directory created on startup if necessary, don't check it
        # exists
        rtdir = abspath(os.environ.get("CW_RUNTIME_DIR", default))
        return join(rtdir, f"{self.appid}-{self.name}.pid")

    # config -> repository

    def repository(self, vreg=None):
        """Return a new bootstrapped repository."""
        from cubicweb.server.repository import Repository

        repo = Repository(self, vreg=vreg)
        repo.bootstrap()
        return repo

    # instance methods used to get instance specific resources #############

    @property
    def appdatahome(self):
        if self.mode == "system":
            if _USR_INSTALL:
                iddir = os.path.join("/var", "lib", "cubicweb", "instances")
            else:
                iddir = os.path.join(
                    _INSTALL_PREFIX, "var", "lib", "cubicweb", "instances"
                )
        else:
            iddir = self.instances_dir()
        iddir = abspath(os.environ.get("CW_INSTANCES_DATA_DIR", iddir))
        return join(iddir, self.appid)

    def add_cubes(self, cubes):
        """add given cubes to the list of used cubes"""
        if not isinstance(cubes, list):
            cubes = list(cubes)
        self._cubes = self.reorder_cubes(list(self._cubes) + cubes)
        self.load_site_cubicweb(cubes)

    def main_config_file(self):
        """return instance's control configuration file"""
        return join(self.apphome, f"{self.name}.conf")

    def save(self):
        """write down current configuration"""
        with open(self.main_config_file(), "w") as fobj:
            self.generate_config(fobj, header_message=CONFIG_HELP_MESSAGE)

    def check_writeable_uid_directory(self, path):
        """check given directory path exists, belongs to the user running the
        server process and is writeable.

        If not, try to fix this, letting exception propagate when not possible.
        """
        if not exists(path):
            self.info("creating %s directory", path)
            try:
                os.makedirs(path)
            except OSError as ex:
                self.warning("error while creating %s directory: %s", path, ex)
                return
        self.ensure_uid(path)

    def get_uid(self):
        if self["uid"]:
            try:
                uid = int(self["uid"])
            except ValueError:
                from pwd import getpwnam

                uid = getpwnam(self["uid"]).pw_uid
        else:
            uid = os.getuid()
        return uid

    def ensure_uid(self, path, enforce_write=False):
        if not exists(path):
            return
        uid = self.get_uid()
        if uid is None:
            return
        fstat = os.stat(path)
        if fstat.st_uid != uid:
            self.info("giving ownership of %s to %s", path, self["uid"])
            try:
                os.chown(path, uid, os.getgid())
            except OSError as ex:
                self.warning(
                    "error while giving ownership of %s to %s: %s",
                    path,
                    self["uid"],
                    ex,
                )

        if enforce_write and not (fstat.st_mode & stat.S_IWUSR):
            self.info("forcing write permission on %s", path)
            try:
                os.chmod(path, fstat.st_mode | stat.S_IWUSR)
            except OSError as ex:
                self.warning("error while forcing write permission on %s: %s", path, ex)

    def ensure_uid_directory(self, path, enforce_write=False):
        self.check_writeable_uid_directory(path)
        for dirpath, dirnames, filenames in os.walk(path):
            for name in filenames:
                self.ensure_uid(join(dirpath, name), enforce_write)
        return path

    @cached
    def instance_md5_version(self):
        from hashlib import md5  # pylint: disable=E0611

        infos = []
        for pkg in sorted(self.cubes()):
            version = self.cube_version(pkg)
            infos.append(f"{pkg}-{version}")
        infos.append(f"cubicweb-{str(self.cubicweb_version())}")
        return md5((";".join(infos)).encode("ascii")).hexdigest()

    def load_configuration(self, **kw):
        """load instance's configuration files"""
        super(CubicWebConfiguration, self).load_configuration(**kw)
        if self.apphome and not self.creating:
            # init gettext
            self._gettext_init()

        self._init_base_url()

    def _generate_base_url(self):
        # normalize base url(s)
        base_url = self["base-url"] or self.default_base_url()

        if base_url and base_url[-1] != "/":
            base_url += "/"

        return base_url

    def _init_base_url(self):
        if not (self.repairing or self.creating):
            self.global_set_option("base-url", self._generate_base_url())

    def default_base_url(self):
        from socket import getfqdn

        return f"http://{self['host'] or getfqdn().lower()}:{self['port'] or 8080}/"

    def available_languages(self, *args):
        """return available translation for an instance, by looking for
        compiled catalog

        take *args to be usable as a vocabulary method
        """
        from glob import glob

        yield "en"  # ensure 'en' is yielded even if no .mo found
        for path in glob(join(self.apphome, "i18n", "*", "LC_MESSAGES")):
            lang = path.split(os.sep)[-2]
            if lang != "en":
                yield lang

    def _gettext_init(self):
        """set language for gettext"""
        from cubicweb.cwgettext import translation

        path = join(self.apphome, "i18n")
        for language in self.available_languages():
            self.info("loading language %s", language)
            try:
                tr = translation("cubicweb", path, languages=[language])
                self.translations[language] = (tr.ugettext, tr.upgettext)
            except IOError:
                if self.mode != "test":
                    # in test contexts, data/i18n does not exist, hence
                    # logging will only pollute the logs
                    self.exception(
                        "localisation support error for language %s", language
                    )

    @staticmethod
    def _sorted_appobjects(appobjects):
        appobjects = sorted(appobjects)
        try:
            index = appobjects.index("entities")
        except ValueError:
            pass
        else:
            # put entities first
            appobjects.insert(0, appobjects.pop(index))
        return appobjects

    def appobjects_cube_modnames(self, cube):
        modnames = []
        cube_modname = _cube_pkgname(cube)
        cube_submodnames = self._sorted_appobjects(self.cube_appobject_path)
        for name in cube_submodnames:
            for modname, filepath in _expand_modname(".".join([cube_modname, name])):
                modnames.append(modname)
        return modnames

    sources_mode = None

    def set_sources_mode(self, sources):
        self.sources_mode = sources

    def i18ncompile(self, langs=None):
        from cubicweb import i18n

        if langs is None:
            langs = self.available_languages()
        i18ndir = join(self.apphome, "i18n")
        if not exists(i18ndir):
            create_dir(i18ndir)
        sourcedirs = [join(path, "i18n") for path in self.cubes_path()]
        sourcedirs.append(self.i18n_lib_dir())
        return i18n.compile_i18n_catalogs(sourcedirs, i18ndir, langs)

    def sendmails(self, msgs, fromaddr=None):
        """msgs: list of 2-uple (message object, recipients). Return False
        if connection to the smtp server failed, else True.
        """
        server, port = self["smtp-host"], self["smtp-port"]
        smtp_user, smtp_password = self["smtp-username"], self["smtp-password"]
        if fromaddr is None:
            fromaddr = f"{self['sender-name']} <{self['sender-addr']}>"
        SMTP_LOCK.acquire()
        try:
            try:
                smtp = self.smtp_cls(server, port)
                try:
                    if smtp_user and smtp_password:
                        smtp.login(smtp_user, smtp_password)
                except SMTPAuthenticationError as exception:
                    self.exception(
                        "cannot log in to SMTP server %s:%s (%s)",
                        server,
                        port,
                        exception,
                    )
            except Exception as ex:
                self.exception(
                    "can't connect to smtp server %s:%s (%s)", server, port, ex
                )
                if self.mode == "test":
                    raise
                return False
            for msg, recipients in msgs:
                try:
                    smtp.sendmail(msg.get("From", fromaddr), recipients, msg.as_bytes())
                except Exception as ex:
                    self.exception("error sending mail to %s (%s)", recipients, ex)
                    if self.mode == "test":
                        raise
            smtp.close()
        finally:
            SMTP_LOCK.release()
        return True

    info = warning = error = critical = exception = debug

    def bootstrap_cubes(self):
        from logilab.common.textutils import splitstrip

        with open(join(self.apphome, "bootstrap_cubes")) as f:
            for line in f:
                line = line.strip()
                if not line or line.startswith("#"):
                    continue
                self.init_cubes(self.expand_cubes(splitstrip(line)))
                break
            else:
                # no cubes
                self.init_cubes(())

    def write_bootstrap_cubes_file(self, cubes):
        stream = open(join(self.apphome, "bootstrap_cubes"), "w")
        stream.write("# this is a generated file only used for bootstraping\n")
        stream.write("# you should not have to edit this\n")
        stream.write(f"{','.join(cubes)}\n")
        stream.close()

    def sources_file(self):
        return join(self.apphome, "sources")

    # this method has to be cached since when the server is running using a
    # restricted user, this user usually don't have access to the sources
    # configuration file (#16102)
    @cached
    def read_sources_file(self):
        """return a dictionary of values found in the sources file"""
        return read_config(self.sources_file(), raise_if_unreadable=True)

    def source_enabled(self, source):
        if self.sources_mode is not None:
            if "migration" in self.sources_mode:
                assert len(self.sources_mode) == 1
                if source.connect_for_migration:
                    return True
                print("not connecting to source", source.uri, "during migration")
                return False
            if "all" in self.sources_mode:
                assert len(self.sources_mode) == 1
                return True
            return source.uri in self.sources_mode
        if self.quick_start:
            return source.uri == "system"
        return not source.disabled and (
            not self.enabled_sources or source.uri in self.enabled_sources
        )

    def write_sources_file(self, sourcescfg):
        """serialize repository'sources configuration into a INI like file"""
        sourcesfile = self.sources_file()
        if exists(sourcesfile):
            import shutil

            shutil.copy(sourcesfile, sourcesfile + ".bak")
        stream = open(sourcesfile, "w")
        stream.write(CONFIG_HELP_MESSAGE)
        for section in ("admin", "system"):
            sconfig = sourcescfg[section]
            if isinstance(sconfig, dict):
                # get a Configuration object
                assert section == "system", f"{section!r} is not system"
                _sconfig = SourceConfiguration(
                    self, options=SOURCE_TYPES["native"].options
                )
                for attr, val in sconfig.items():
                    try:
                        _sconfig.set_option(attr, val)
                    except OptionError:
                        # skip adapter, may be present on pre 3.10 instances
                        if attr != "adapter":
                            self.error(f"skip unknown option {attr} in sources file")
                sconfig = _sconfig
            stream.write(f"[{section}]\n{generate_source_config(sconfig)}\n")
        restrict_perms_to_user(sourcesfile)

    def load_schema(self, expand_cubes=False, **kwargs):
        from cubicweb.schema import CubicWebSchemaLoader

        if expand_cubes:
            # in case some new dependencies have been introduced, we have to
            # reinitialize cubes so the full filesystem schema is read
            origcubes = self.cubes()
            self._cubes = None
            self.init_cubes(self.expand_cubes(origcubes))
        schema = CubicWebSchemaLoader().load(self, **kwargs)
        if expand_cubes:
            # restore original value
            self._cubes = origcubes
        return schema

    def load_bootstrap_schema(self):
        from cubicweb.schema import BootstrapSchemaLoader

        schema = BootstrapSchemaLoader().load(self)
        schema.name = "bootstrap"
        return schema

    def migration_handler(
        self,
        schema=None,
        interactive=True,
        cnx=None,
        repo=None,
        connect=True,
        verbosity=None,
    ):
        """return a migration handler instance"""
        from cubicweb.server.migractions import ServerMigrationHelper

        if verbosity is None:
            verbosity = getattr(self, "verbosity", 0)
        return ServerMigrationHelper(
            self,
            schema,
            interactive=interactive,
            cnx=cnx,
            repo=repo,
            connect=connect,
            verbosity=verbosity,
        )

    def anonymous_user(self):
        """return a login and password to use for anonymous users.

        None may be returned for both if anonymous connection is not
        allowed or if an empty login is used in configuration
        """
        try:
            user = self["anonymous-user"] or None
            passwd = self["anonymous-password"]
        except KeyError:
            user, passwd = None, None
        except UnicodeDecodeError:
            raise ConfigurationError("anonymous information should only contains ascii")
        return user, passwd

    @property
    def system_source_config(self):
        return self.read_sources_file()["system"]

    @property
    def default_admin_config(self):
        return self.read_sources_file()["admin"]


set_log_methods(CubicWebConfiguration, logging.getLogger("cubicweb.configuration"))

# alias to get a configuration instance from an instance id
instance_configuration = CubicWebConfiguration.config_for


_EXT_REGISTERED = False


def register_stored_procedures():
    from logilab.database import FunctionDescr
    from rql.utils import register_function, iter_funcnode_variables
    from rql.nodes import SortTerm, Constant, VariableRef

    global _EXT_REGISTERED
    if _EXT_REGISTERED:
        return
    _EXT_REGISTERED = True

    class COMMA_JOIN(FunctionDescr):
        supported_backends = (
            "postgres",
            "sqlite",
        )
        rtype = "String"

        def st_description(self, funcnode, mainindex, tr):
            return ", ".join(
                sorted(
                    term.get_description(mainindex, tr)
                    for term in iter_funcnode_variables(funcnode)
                )
            )

    register_function(COMMA_JOIN)  # XXX do not expose?

    class CONCAT_STRINGS(COMMA_JOIN):
        aggregat = True

    register_function(CONCAT_STRINGS)  # XXX bw compat

    class GROUP_CONCAT(CONCAT_STRINGS):
        supported_backends = (
            "postgres",
            "sqlite",
        )

    register_function(GROUP_CONCAT)

    class GROUP_LIST(FunctionDescr):
        aggregat = True
        supported_backends = ("postgres",)
        minargs = maxargs = 1
        rtype = "List[Any]"

        def as_sql_postgres(self, args):
            return f"ARRAY_AGG({args[0]})"

    register_function(GROUP_LIST)

    class LIMIT_SIZE(FunctionDescr):
        supported_backends = (
            "postgres",
            "sqlite",
        )
        minargs = maxargs = 3
        rtype = "String"

        def st_description(self, funcnode, mainindex, tr):
            return funcnode.children[0].get_description(mainindex, tr)

    register_function(LIMIT_SIZE)

    class TEXT_LIMIT_SIZE(LIMIT_SIZE):
        supported_backends = (
            "postgres",
            "sqlite",
        )
        minargs = maxargs = 2

    register_function(TEXT_LIMIT_SIZE)

    class FTIRANK(FunctionDescr):
        """return ranking of a variable that must be used as some has_text
        relation subject in the query's restriction. Usually used to sort result
        of full-text search by ranking.
        """

        supported_backends = ("postgres",)
        rtype = "Float"

        def st_check_backend(self, backend, funcnode):
            """overriden so that on backend not supporting fti ranking, the
            function is removed when in an orderby clause, or replaced by a 1.0
            constant.
            """
            if not self.supports(backend):
                parent = funcnode.parent
                while parent is not None and not isinstance(parent, SortTerm):
                    parent = parent.parent
                if isinstance(parent, SortTerm):
                    parent.parent.remove(parent)
                else:
                    funcnode.parent.replace(funcnode, Constant(1.0, "Float"))
                    parent = funcnode
                for vref in parent.iget_nodes(VariableRef):
                    vref.unregister_reference()

    register_function(FTIRANK)

    class FTIHL(FunctionDescr):
        """return fragments of the document with marked search terms

        eg: Any X, FTIHL(C, "hello") WHERE X has_text "hello", X content C
        eg: Any X, FTIHL(C, "hello", "MaxFragments=2") WHERE X has_text "hello", X content C

        options are comma separated.
        See https://www.postgresql.org/docs/current/textsearch-controls.html#TEXTSEARCH-HEADLINE
        for the list of available options
        """

        supported_backends = ("postgres",)
        rtype = "String"
        minargs = 2
        maxargs = 3

        def as_sql_postgres(self, args):
            attr, search_text = args[:2]
            options = args[2] if len(args) == 3 else None

            if not options:
                return f"ts_headline({attr}, plainto_tsquery({search_text}))"
            return f"ts_headline({attr}, plainto_tsquery({search_text}), {options})"

    register_function(FTIHL)

    class FSPATH(FunctionDescr):
        """return path of some bytes attribute stored using the Bytes
        File-System Storage (bfss)
        """

        rtype = "Bytes"  # XXX return a String? potential pb with fs encoding

        def update_cb_stack(self, stack):
            assert len(stack) == 1
            stack[0] = self.source_execute

        def as_sql(self, backend, args):
            raise NotImplementedError(
                "This callback is only available for BytesFileSystemStorage "
                "managed attribute. Is FSPATH() argument BFSS managed?"
            )

        def source_execute(self, source, session, value):
            fpath = source.binary_to_str(value)
            try:
                return Binary(fpath)
            except OSError as ex:
                source.critical("can't open %s: %s", fpath, ex)
                return None

    register_function(FSPATH)



# copyright 2003-2010 LOGILAB S.A. (Paris, FRANCE), all rights reserved.
# contact https://www.logilab.fr/ -- mailto:contact@logilab.fr
#
# This file is part of CubicWeb.
#
# CubicWeb is free software: you can redistribute it and/or modify it under the
# terms of the GNU Lesser General Public License as published by the Free
# Software Foundation, either version 2.1 of the License, or (at your option)
# any later version.
#
# CubicWeb is distributed in the hope that it will be useful, but WITHOUT
# ANY WARRANTY; without even the implied warranty of MERCHANTABILITY or FITNESS
# FOR A PARTICULAR PURPOSE.  See the GNU Lesser General Public License for more
# details.
#
# You should have received a copy of the GNU Lesser General Public License along
# with CubicWeb.  If not, see <https://www.gnu.org/licenses/>.
"""map standard cubicweb schema to xml vocabularies"""

from yams import xy

xy.register_prefix("rdf", "http://www.w3.org/1999/02/22-rdf-syntax-ns#")
xy.register_prefix("dc", "http://purl.org/dc/elements/1.1/")
xy.register_prefix("foaf", "http://xmlns.com/foaf/0.1/")
xy.register_prefix("doap", "http://usefulinc.com/ns/doap#")
xy.register_prefix("owl", "http://www.w3.org/2002/07/owl#")
xy.register_prefix("dcterms", "http://purl.org/dc/terms/")

xy.add_equivalence("creation_date", "dc:date")
xy.add_equivalence("created_by", "dc:creator")
xy.add_equivalence("description", "dc:description")
xy.add_equivalence("CWUser", "foaf:Person")
xy.add_equivalence("CWUser login", "foaf:Person dc:title")
xy.add_equivalence("CWUser surname", "foaf:Person foaf:name")



# copyright 2003-2016 LOGILAB S.A. (Paris, FRANCE), all rights reserved.
# contact https://www.logilab.fr/ -- mailto:contact@logilab.fr
#
# This file is part of CubicWeb.
#
# CubicWeb is free software: you can redistribute it and/or modify it under the
# terms of the GNU Lesser General Public License as published by the Free
# Software Foundation, either version 2.1 of the License, or (at your option)
# any later version.
#
# CubicWeb is distributed in the hope that it will be useful, but WITHOUT
# ANY WARRANTY; without even the implied warranty of MERCHANTABILITY or FITNESS
# FOR A PARTICULAR PURPOSE.  See the GNU Lesser General Public License for more
# details.
#
# You should have received a copy of the GNU Lesser General Public License along
# with CubicWeb.  If not, see <https://www.gnu.org/licenses/>.
"""unit/functional tests for cubicweb.server.hook"""

import unittest

from logilab.common.testlib import mock_object

from cubicweb.devtools import fake
from cubicweb.devtools.apptest_config import ApptestConfiguration
from cubicweb.devtools.testlib import CubicWebTC
from cubicweb.hooks import integrity, syncschema
from cubicweb.server import hook


class OperationsTC(CubicWebTC):
    def test_late_operation(self):
        with self.admin_access.repo_cnx() as cnx:
            l1 = hook.LateOperation(cnx)
            l2 = hook.LateOperation(cnx)
            l3 = hook.Operation(cnx)
            self.assertEqual(cnx.pending_operations, [l3, l1, l2])

    def test_single_last_operation(self):
        with self.admin_access.repo_cnx() as cnx:
            l0 = hook.SingleLastOperation(cnx)
            l1 = hook.LateOperation(cnx)
            l2 = hook.LateOperation(cnx)
            l3 = hook.Operation(cnx)
            self.assertEqual(cnx.pending_operations, [l3, l1, l2, l0])
            l4 = hook.SingleLastOperation(cnx)
            self.assertEqual(cnx.pending_operations, [l3, l1, l2, l4])

    def test_global_operation_order(self):
        with self.admin_access.repo_cnx() as cnx:
            op1 = syncschema.RDefDelOp(cnx)
            op2 = integrity._CheckORelationOp(cnx)
            op3 = syncschema.MemSchemaNotifyChanges(cnx)
            self.assertEqual([op1, op2, op3], cnx.pending_operations)


class HookCalled(Exception):
    pass


def setUpModule():
    global config, schema
    config = ApptestConfiguration("data", __file__)
    config.bootstrap_cubes()
    schema = config.load_schema()


def tearDownModule(*args):
    global config, schema
    del config, schema


class AddAnyHook(hook.Hook):
    __regid__ = "addany"
    category = "cat1"
    events = ("before_add_entity",)

    def __call__(self):
        raise HookCalled()


class HooksRegistryTC(unittest.TestCase):
    def setUp(self):
        """called before each test from this class"""
        self.vreg = mock_object(config=config, schema=schema)
        self.o = hook.HooksRegistry(self.vreg)

    def test_register_bad_hook1(self):
        class _Hook(hook.Hook):
            events = ("before_add_entiti",)

        with self.assertRaises(Exception) as cm:
            self.o.register(_Hook)
        self.assertEqual(
            str(cm.exception), f"bad event before_add_entiti on {__name__}._Hook"
        )

    def test_register_bad_hook2(self):
        class _Hook(hook.Hook):
            events = None

        with self.assertRaises(Exception) as cm:
            self.o.register(_Hook)
        self.assertEqual(
            str(cm.exception), f"bad .events attribute None on {__name__}._Hook"
        )

    def test_register_bad_hook3(self):
        class _Hook(hook.Hook):
            events = "before_add_entity"

        with self.assertRaises(Exception) as cm:
            self.o.register(_Hook)
        self.assertEqual(str(cm.exception), f"bad event b on {__name__}._Hook")

    def test_call_hook(self):
        self.o.register(AddAnyHook)
        active_hooks = set(["cat1"])
        cw = fake.FakeConnection()
        cw.is_hook_activated = lambda cls: cls.category in active_hooks
        # active hook is called
        self.assertRaises(HookCalled, self.o.call_hooks, "before_add_entity", cw)
        # inactive hook is not called
        active_hooks.remove("cat1")
        self.o.call_hooks("before_add_entity", cw)
        # reactivated hook is called
        active_hooks.add("cat1")
        self.assertRaises(HookCalled, self.o.call_hooks, "before_add_entity", cw)
        # unregisterd hook is not called
        self.o.unregister(AddAnyHook)
        self.o.call_hooks("before_add_entity", cw)


class SystemHooksTC(CubicWebTC):
    def test_startup_shutdown(self):
        import hooks  # cubicweb/server/test/data/hooks.py

        self.assertEqual(hooks.CALLED_EVENTS["server_startup"], True)
        # don't actually call repository.shutdown !
        self.repo.hm.call_hooks("server_shutdown", repo=self.repo)
        self.assertEqual(hooks.CALLED_EVENTS["server_shutdown"], True)


if __name__ == "__main__":
    unittest.main()



# copyright 2003-2013 LOGILAB S.A. (Paris, FRANCE), all rights reserved.
# contact https://www.logilab.fr/ -- mailto:contact@logilab.fr
#
# This file is part of CubicWeb.
#
# CubicWeb is free software: you can redistribute it and/or modify it under the
# terms of the GNU Lesser General Public License as published by the Free
# Software Foundation, either version 2.1 of the License, or (at your option)
# any later version.
#
# CubicWeb is distributed in the hope that it will be useful, but WITHOUT
# ANY WARRANTY; without even the implied warranty of MERCHANTABILITY or FITNESS
# FOR A PARTICULAR PURPOSE.  See the GNU Lesser General Public License for more
# details.
#
# You should have received a copy of the GNU Lesser General Public License along
# with CubicWeb.  If not, see <https://www.gnu.org/licenses/>.
"""core CubicWeb schema necessary for bootstrapping the actual instance's schema"""


from yams.buildobjs import (
    EntityType,
    RelationType,
    RelationDefinition,
    Bytes,
    SubjectRelation,
    RichString,
    String,
    Boolean,
    Int,
)

from cubicweb import _
from cubicweb.schema import (
    RQLConstraint,
    PUB_SYSTEM_ENTITY_PERMS,
    PUB_SYSTEM_REL_PERMS,
    PUB_SYSTEM_ATTR_PERMS,
)


# not restricted since as "is" is handled as other relations, guests need
# access to this


class CWEType(EntityType):
    """define an entity type, used to build the instance schema"""

    __permissions__ = PUB_SYSTEM_ENTITY_PERMS
    name = String(
        required=True, indexed=True, internationalizable=True, unique=True, maxsize=64
    )
    description = RichString(
        internationalizable=True,
        description=_("semantic description of this entity type"),
    )
    # necessary to filter using RQL
    final = Boolean(default=False, description=_("automatic"))


class CWRType(EntityType):
    """define a relation type, used to build the instance schema"""

    __permissions__ = PUB_SYSTEM_ENTITY_PERMS
    name = String(
        required=True, indexed=True, internationalizable=True, unique=True, maxsize=64
    )
    description = RichString(
        internationalizable=True,
        description=_("semantic description of this relation type"),
    )
    symmetric = Boolean(
        description=_("is this relation equivalent in both direction ?")
    )
    inlined = Boolean(
        description=_(
            "is this relation physically inlined? "
            "you should know what you're doing if you are changing this!"
        )
    )
    fulltext_container = String(
        description=_(
            "if full text content of subject/object entity "
            "should be added to other side entity (the container)."
        ),
        vocabulary=("", _("subject"), _("object")),
        maxsize=8,
        default=None,
    )
    final = Boolean(description=_("automatic"))


class CWComputedRType(EntityType):
    """define a virtual relation type, used to build the instance schema"""

    __permissions__ = PUB_SYSTEM_ENTITY_PERMS
    name = String(
        required=True, indexed=True, internationalizable=True, unique=True, maxsize=64
    )
    description = RichString(
        internationalizable=True,
        description=_("semantic description of this relation type"),
    )
    rule = String(required=True)


class CWAttribute(EntityType):
    """define a final relation: link a final relation type from a non final
    entity to a final entity type.

    used to build the instance schema
    """

    __permissions__ = PUB_SYSTEM_ENTITY_PERMS
    relation_type = SubjectRelation(
        "CWRType",
        cardinality="1*",
        constraints=[RQLConstraint("O final TRUE")],
        composite="object",
    )
    from_entity = SubjectRelation(
        "CWEType",
        cardinality="1*",
        constraints=[RQLConstraint("O final FALSE")],
        composite="object",
    )
    to_entity = SubjectRelation(
        "CWEType",
        cardinality="1*",
        constraints=[RQLConstraint("O final TRUE")],
        composite="object",
    )
    constrained_by = SubjectRelation(
        "CWConstraint", cardinality="*1", composite="subject"
    )

    cardinality = String(
        maxsize=2,
        internationalizable=True,
        vocabulary=[_("?1"), _("11")],
        description=_("subject/object cardinality"),
    )
    ordernum = Int(description=("control subject entity's relations order"), default=0)

    formula = String(maxsize=2048)
    indexed = Boolean(
        description=_("create an index for quick search on this attribute")
    )
    fulltextindexed = Boolean(
        description=_("index this attribute's value in the plain text index")
    )
    internationalizable = Boolean(
        description=_("is this attribute's value translatable")
    )
    defaultval = Bytes(description=_("default value as gziped pickled python object"))
    extra_props = Bytes(description=_("additional type specific properties"))

    description = RichString(
        internationalizable=True,
        description=_("semantic description of this attribute"),
    )


CARDINALITY_VOCAB = [
    _("?*"),
    _("1*"),
    _("+*"),
    _("**"),
    _("?+"),
    _("1+"),
    _("++"),
    _("*+"),
    _("?1"),
    _("11"),
    _("+1"),
    _("*1"),
    _("??"),
    _("1?"),
    _("+?"),
    _("*?"),
]


class CWRelation(EntityType):
    """define a non final relation: link a non final relation type from a non
    final entity to a non final entity type.

    used to build the instance schema
    """

    __permissions__ = PUB_SYSTEM_ENTITY_PERMS
    relation_type = SubjectRelation(
        "CWRType",
        cardinality="1*",
        constraints=[RQLConstraint("O final FALSE")],
        composite="object",
    )
    from_entity = SubjectRelation(
        "CWEType",
        cardinality="1*",
        constraints=[RQLConstraint("O final FALSE")],
        composite="object",
    )
    to_entity = SubjectRelation(
        "CWEType",
        cardinality="1*",
        constraints=[RQLConstraint("O final FALSE")],
        composite="object",
    )
    constrained_by = SubjectRelation(
        "CWConstraint", cardinality="*1", composite="subject"
    )

    cardinality = String(
        maxsize=2,
        internationalizable=True,
        vocabulary=CARDINALITY_VOCAB,
        description=_("subject/object cardinality"),
    )
    ordernum = Int(description=_("control subject entity's relations order"), default=0)
    composite = String(
        description=_(
            "is the subject/object entity of the relation "
            "composed of the other ? This implies that when "
            "the composite is deleted, composants are also "
            "deleted."
        ),
        vocabulary=("", _("subject"), _("object")),
        maxsize=8,
        default=None,
    )

    description = RichString(
        internationalizable=True, description=_("semantic description of this relation")
    )


# not restricted since it has to be read when checking allowed transitions
class RQLExpression(EntityType):
    """define a rql expression used to define permissions"""

    __permissions__ = PUB_SYSTEM_ENTITY_PERMS
    exprtype = String(required=True, vocabulary=["ERQLExpression", "RRQLExpression"])
    mainvars = String(
        maxsize=8,
        description=_(
            "name of the main variables which should be "
            "used in the selection if necessary (comma "
            "separated)"
        ),
    )
    expression = String(
        required=True,
        description=_(
            "restriction part of a rql query. "
            "For entity rql expression, X and U are "
            "predefined respectivly to the current object and to "
            "the request user. For relation rql expression, "
            "S, O and U are predefined respectivly to the current "
            "relation'subject, object and to "
            "the request user. "
        ),
    )


class CWConstraint(EntityType):
    """define a schema constraint"""

    __permissions__ = PUB_SYSTEM_ENTITY_PERMS
    cstrtype = SubjectRelation("CWConstraintType", cardinality="1*")
    value = String(description=_("depends on the constraint type"))


class CWUniqueTogetherConstraint(EntityType):
    """defines a sql-level multicolumn unique index"""

    __permissions__ = PUB_SYSTEM_ENTITY_PERMS
    name = String(required=True, unique=True, maxsize=64)
    constraint_of = SubjectRelation(
        "CWEType", cardinality="1*", composite="object", inlined=True
    )
    relations = SubjectRelation(
        "CWRType",
        cardinality="+*",
        constraints=[
            RQLConstraint(
                "S constraint_of ET, RDEF relation_type O, RDEF from_entity ET, "
                "O final TRUE OR (O final FALSE AND O inlined TRUE)"
            )
        ],
    )


class CWConstraintType(EntityType):
    """define a schema constraint type"""

    __permissions__ = PUB_SYSTEM_ENTITY_PERMS
    name = String(
        required=True, indexed=True, internationalizable=True, unique=True, maxsize=64
    )


# not restricted since it has to be read when checking allowed transitions
class CWGroup(EntityType):
    """define a CubicWeb users group"""

    __permissions__ = PUB_SYSTEM_ENTITY_PERMS
    name = String(
        required=True, indexed=True, internationalizable=True, unique=True, maxsize=64
    )


class CWProperty(EntityType):
    """used for cubicweb configuration. Once a property has been created you
    can't change the key.
    """

    __permissions__ = {
        "read": ("managers", "users", "guests"),
        "add": (
            "managers",
            "users",
        ),
        "update": (
            "managers",
            "owners",
        ),
        "delete": (
            "managers",
            "owners",
        ),
    }
    # key is a reserved word for mysql
    pkey = String(
        required=True,
        internationalizable=True,
        maxsize=256,
        description=_(
            "defines what's the property is applied for. "
            "You must select this first to be able to set "
            "value"
        ),
    )
    value = String(internationalizable=True, maxsize=256)


class relation_type(RelationType):
    """link a relation definition to its relation type"""

    __permissions__ = PUB_SYSTEM_REL_PERMS
    inlined = True


class from_entity(RelationType):
    """link a relation definition to its subject entity type"""

    __permissions__ = PUB_SYSTEM_REL_PERMS
    inlined = True


class to_entity(RelationType):
    """link a relation definition to its object entity type"""

    __permissions__ = PUB_SYSTEM_REL_PERMS
    inlined = True


class constrained_by(RelationType):
    """constraints applying on this relation"""

    __permissions__ = PUB_SYSTEM_REL_PERMS


class cstrtype(RelationType):
    """constraint factory"""

    __permissions__ = PUB_SYSTEM_REL_PERMS
    inlined = True


class read_permission_cwgroup(RelationDefinition):
    """groups allowed to read entities/relations of this type"""

    __permissions__ = PUB_SYSTEM_REL_PERMS
    name = "read_permission"
    subject = ("CWEType", "CWAttribute", "CWRelation", "CWComputedRType")
    object = "CWGroup"
    cardinality = "**"


class add_permission_cwgroup(RelationDefinition):
    """groups allowed to add entities/relations of this type"""

    __permissions__ = PUB_SYSTEM_REL_PERMS
    name = "add_permission"
    subject = ("CWEType", "CWRelation", "CWAttribute")
    object = "CWGroup"
    cardinality = "**"


class delete_permission_cwgroup(RelationDefinition):
    """groups allowed to delete entities/relations of this type"""

    __permissions__ = PUB_SYSTEM_REL_PERMS
    name = "delete_permission"
    subject = ("CWEType", "CWRelation")
    object = "CWGroup"
    cardinality = "**"


class update_permission_cwgroup(RelationDefinition):
    """groups allowed to update entities/relations of this type"""

    __permissions__ = PUB_SYSTEM_REL_PERMS
    name = "update_permission"
    subject = ("CWEType", "CWAttribute")
    object = "CWGroup"
    cardinality = "**"


class read_permission_rqlexpr(RelationDefinition):
    """rql expression allowing to read entities/relations of this type"""

    __permissions__ = PUB_SYSTEM_REL_PERMS
    name = "read_permission"
    subject = ("CWEType", "CWAttribute", "CWRelation", "CWComputedRType")
    object = "RQLExpression"
    cardinality = "*?"
    composite = "subject"


class add_permission_rqlexpr(RelationDefinition):
    """rql expression allowing to add entities/relations of this type"""

    __permissions__ = PUB_SYSTEM_REL_PERMS
    name = "add_permission"
    subject = ("CWEType", "CWRelation", "CWAttribute")
    object = "RQLExpression"
    cardinality = "*?"
    composite = "subject"


class delete_permission_rqlexpr(RelationDefinition):
    """rql expression allowing to delete entities/relations of this type"""

    __permissions__ = PUB_SYSTEM_REL_PERMS
    name = "delete_permission"
    subject = ("CWEType", "CWRelation")
    object = "RQLExpression"
    cardinality = "*?"
    composite = "subject"


class update_permission_rqlexpr(RelationDefinition):
    """rql expression allowing to update entities/relations of this type"""

    __permissions__ = PUB_SYSTEM_REL_PERMS
    name = "update_permission"
    subject = ("CWEType", "CWAttribute")
    object = "RQLExpression"
    cardinality = "*?"
    composite = "subject"


class is_(RelationType):
    """core relation indicating the type of an entity"""

    name = "is"
    # don't explicitly set composite here, this is handled anyway
    # composite = 'object'
    __permissions__ = {
        "read": ("managers", "users", "guests"),
        "add": (),
        "delete": (),
    }
    cardinality = "1*"
    subject = "*"
    object = "CWEType"


class is_instance_of(RelationType):
    """core relation indicating the types (including specialized types)
    of an entity
    """

    # don't explicitly set composite here, this is handled anyway
    # composite = 'object'
    __permissions__ = {
        "read": ("managers", "users", "guests"),
        "add": (),
        "delete": (),
    }
    cardinality = "+*"
    subject = "*"
    object = "CWEType"


class specializes(RelationType):
    name = "specializes"
    __permissions__ = {
        "read": ("managers", "users", "guests"),
        "add": ("managers",),
        "delete": ("managers",),
    }
    cardinality = "?*"
    subject = "CWEType"
    object = "CWEType"


def post_build_callback(schema):
    """set attributes permissions for schema/workflow entities"""
    from cubicweb.schema import SCHEMA_TYPES, WORKFLOW_TYPES, META_RTYPES

    wftypes = WORKFLOW_TYPES - {"TrInfo"}
    for eschema in schema.entities():
        if eschema in SCHEMA_TYPES or eschema in wftypes:
            for rschema in list(eschema.subject_relations.values()):
                if rschema.final and rschema not in META_RTYPES:
                    rdef = eschema.relation_definition(rschema)
                    rdef.permissions = PUB_SYSTEM_ATTR_PERMS



# copyright 2003-2013 LOGILAB S.A. (Paris, FRANCE), all rights reserved.
# contact https://www.logilab.fr/ -- mailto:contact@logilab.fr
#
# This file is part of CubicWeb.
#
# CubicWeb is free software: you can redistribute it and/or modify it under the
# terms of the GNU Lesser General Public License as published by the Free
# Software Foundation, either version 2.1 of the License, or (at your option)
# any later version.
#
# CubicWeb is distributed in the hope that it will be useful, but WITHOUT
# ANY WARRANTY; without even the implied warranty of MERCHANTABILITY or FITNESS
# FOR A PARTICULAR PURPOSE.  See the GNU Lesser General Public License for more
# details.
#
# You should have received a copy of the GNU Lesser General Public License along
# with CubicWeb.  If not, see <https://www.gnu.org/licenses/>.
"""unittests for cubicweb.uilib"""


import doctest
from unittest import skipIf

import sys

if sys.version_info < (3, 10):
    from importlib_metadata import version
else:
    from importlib.metadata import version

from logilab.common.testlib import TestCase, unittest_main

from cubicweb import uilib, utils as cwutils

lxml_version = version("lxml").split(".")


class UILIBTC(TestCase):
    def test_remove_tags(self):
        """make sure remove_tags remove all tags"""
        data = [
            ("<h1>Hello</h1>", "Hello"),
            ('<h1>Hello <a href="foo/bar"><b>s</b>pam</a></h1>', "Hello spam"),
            ('<br>Hello<img src="doh.png"/>', "Hello"),
            ("<p></p>", ""),
        ]
        for text, expected in data:
            got = uilib.remove_html_tags(text)
            self.assertEqual(got, expected)

    def test_lxml_safe_cut(self):
        self.assertEqual(
            uilib.safe_cut("aaa<div>aaad</div> ef", 4), "<p>aaa</p><div>a...</div>"
        )
        self.assertEqual(
            uilib.safe_cut("aaa<div>aaad</div> ef", 7), "<p>aaa</p><div>aaad</div>..."
        )
        self.assertEqual(
            uilib.safe_cut("aaa<div>aaad</div>", 7), "<p>aaa</p><div>aaad</div>"
        )
        # Missing ellipsis due to space management but we don't care
        self.assertEqual(
            uilib.safe_cut('ab <a href="hello">&amp;d</a>', 4),
            '<p>ab <a href="hello">&amp;...</a></p>',
        )

    def test_cut(self):
        """tests uilib.cut() behaviour"""
        data = [
            ("hello", "hello"),
            ("hello world", "hello wo..."),
            ("hell<b>O'</b> world", "hell<b>O..."),
        ]
        for text, expected in data:
            got = uilib.cut(text, 8)
            self.assertEqual(got, expected)

    def test_text_cut(self):
        """tests uilib.text_cut() behaviour with no text"""
        data = [
            ("", ""),
            (
                """Lorem ipsum dolor sit amet, consectetur adipisicing elit, sed do eiusmod
tempor incididunt ut labore et dolore magna aliqua. Ut enim ad minim veniam,
quis nostrud exercitation ullamco laboris nisi ut aliquip ex ea commodo
consequat. Duis aute irure dolor in reprehenderit in voluptate velit esse
cillum dolore eu fugiat nulla pariatur.""",
                "Lorem ipsum dolor sit amet, consectetur adipisicing elit, sed do eiusmod \
tempor incididunt ut labore et dolore magna aliqua. Ut enim ad minim veniam, \
quis nostrud exercitation ullamco laboris nisi ut aliquip ex ea commodo \
consequat.",
            ),
            (
                """Lorem ipsum dolor sit amet, consectetur adipisicing elit, sed do eiusmod
tempor incididunt ut labore et dolore magna aliqua Ut enim ad minim veniam,
quis nostrud exercitation ullamco laboris nisi ut aliquip ex ea commodo
consequat Duis aute irure dolor in reprehenderit in voluptate velit esse
cillum dolore eu fugiat nulla pariatur Excepteur sint occaecat cupidatat non
proident, sunt in culpa qui officia deserunt mollit anim id est laborum
""",
                "Lorem ipsum dolor sit amet, consectetur adipisicing elit, sed do eiusmod \
tempor incididunt ut labore et dolore magna aliqua Ut enim ad minim veniam, \
quis nostrud exercitation ullamco laboris nisi",
            ),
        ]
        for text, expected in data:
            got = uilib.text_cut(text, 30)
            self.assertEqual(got, expected)

    def test_soup2xhtml_0(self):
        self.assertEqual(uilib.soup2xhtml("hop\r\nhop", "ascii"), "hop\nhop")

    def test_soup2xhtml_1_1(self):
        self.assertEqual(uilib.soup2xhtml("hop", "ascii"), "hop")
        self.assertEqual(uilib.soup2xhtml("hop<div>", "ascii"), "hop<div/>")
        self.assertEqual(uilib.soup2xhtml("hop <div>", "ascii"), "hop <div/>")
        self.assertEqual(uilib.soup2xhtml("<div> hop", "ascii"), "<div> hop</div>")
        self.assertEqual(
            uilib.soup2xhtml("hop <div> hop", "ascii"), "hop <div> hop</div>"
        )

    def test_soup2xhtml_1_2(self):
        self.assertEqual(uilib.soup2xhtml("hop </div>", "ascii"), "hop ")
        self.assertEqual(uilib.soup2xhtml("</div> hop", "ascii"), "<div/> hop")
        self.assertEqual(
            uilib.soup2xhtml("hop </div> hop", "ascii"), "<div>hop </div> hop"
        )

    def test_soup2xhtml_2_1(self):
        self.assertEqual(uilib.soup2xhtml("hop <body>", "ascii"), "hop ")
        self.assertEqual(uilib.soup2xhtml("<body> hop", "ascii"), " hop")
        self.assertEqual(uilib.soup2xhtml("hop <body> hop", "ascii"), "hop  hop")

    def test_soup2xhtml_2_2a(self):
        self.assertEqual(uilib.soup2xhtml("hop </body>", "ascii"), "hop ")
        self.assertEqual(uilib.soup2xhtml("</body> hop", "ascii"), " hop")

    @skipIf(
        lxml_version < ["2", "2"], "expected behaviour on recent version of lxml only"
    )
    def test_soup2xhtml_2_2b(self):
        self.assertEqual(uilib.soup2xhtml("hop </body> hop", "ascii"), "hop  hop")

    def test_soup2xhtml_3_1(self):
        self.assertEqual(uilib.soup2xhtml("hop <html>", "ascii"), "hop ")
        self.assertEqual(uilib.soup2xhtml("<html> hop", "ascii"), " hop")
        self.assertEqual(uilib.soup2xhtml("hop <html> hop", "ascii"), "hop  hop")

    def test_soup2xhtml_3_2(self):
        self.assertEqual(uilib.soup2xhtml("hop </html>", "ascii"), "hop ")
        self.assertEqual(uilib.soup2xhtml("</html> hop", "ascii"), " hop")
        self.assertEqual(uilib.soup2xhtml("hop </html> hop", "ascii"), "hop  hop")

    def test_soup2xhtml_3_3(self):
        self.assertEqual(
            uilib.soup2xhtml("<script>test</script> hop ", "ascii"), " hop "
        )

    def test_js(self):
        self.assertEqual(str(uilib.js.pouet(1, "2")), 'pouet(1,"2")')
        self.assertEqual(str(uilib.js.cw.pouet(1, "2")), 'cw.pouet(1,"2")')
        self.assertEqual(
            str(uilib.js.cw.pouet(1, "2").pouet(None)), 'cw.pouet(1,"2").pouet(null)'
        )
        self.assertEqual(
            str(uilib.js.cw.pouet(1, cwutils.JSString("$")).pouet(None)),
            "cw.pouet(1,$).pouet(null)",
        )
        self.assertEqual(
            str(
                uilib.js.cw.pouet(1, {"call back": cwutils.JSString("cw.cb")}).pouet(
                    None
                )
            ),
            'cw.pouet(1,{"call back": cw.cb}).pouet(null)',
        )

    def test_embedded_css(self):
        incoming = """voir le ticket <style type="text/css">@font-face { font-family: "Cambria"; }p.MsoNormal, li.MsoNormal, div.MsoNormal { margin: 0cm 0cm 10pt; font-size: 12pt; font-family: "Times New Roman"; }a:link, span.MsoHyperlink { color: blue; text-decoration: underline; }a:visited, span.MsoHyperlinkFollowed { color: purple; text-decoration: underline; }div.Section1 { page: Section1; }</style></p><p class="MsoNormal">text</p>"""
        expected = 'voir le ticket <p class="MsoNormal">text</p>'
        self.assertMultiLineEqual(uilib.soup2xhtml(incoming, "ascii"), expected)

    def test_unknown_namespace(self):
        incoming = """<table cellspacing="0" cellpadding="0" width="81" border="0" x:str="" style="width: 61pt; border-collapse: collapse">\
<colgroup><col width="81" style="width: 61pt; mso-width-source: userset; mso-width-alt: 2962"/></colgroup>\
<tbody><tr height="17" style="height: 12.75pt"><td width="81" height="17" style="border-right: #e0dfe3; border-top: #e0dfe3; border-left: #e0dfe3; width: 61pt; border-bottom: #e0dfe3; height: 12.75pt; background-color: transparent"><font size="2">XXXXXXX</font></td></tr></tbody>\
</table>"""
        expected = """<table cellspacing="0" cellpadding="0" width="81" border="0">\
<colgroup><col width="81"/></colgroup>\
<tbody><tr height="17"><td width="81" height="17">XXXXXXX</td></tr></tbody>\
</table>"""
        self.assertMultiLineEqual(uilib.soup2xhtml(incoming, "ascii"), expected)


def load_tests(loader, tests, ignore):
    tests.addTests(doctest.DocTestSuite(uilib))
    return tests


if __name__ == "__main__":
    unittest_main()
